import numpy as np


class RegionOfInterest(object):

    def __init__(self, xmin, xmax, ymin, ymax):
        self.xmin= int(xmin)
        self.xmax= int(xmax)
        self.ymin= int(ymin)
        self.ymax= int(ymax)


    def toNumpyArray(self):
        return np.array([self.xmin, self.xmax, self.ymin, self.xmax])


    def __repr__(self):
        return "[%d:%d, %d:%d]" % (self.xmin, self.xmax, self.ymin, self.xmax)


    def cutOut(self, frame):
        return frame[self.ymin: self.ymax, self.xmin: self.xmax]



from arte.utils.zernike_generator import ZernikeGenerator
from arte.types.zernike_coefficients import ZernikeCoefficients
from arte.utils.base_modal_decomposer import BaseModalDecomposer


class ZernikeModalDecomposer(BaseModalDecomposer):
    """
    This class decomposes a wavefront or slope array into a set of modal zernike coefficients.

    Parameters
    ----------
    n_modes: int
        Number of modes to decompose the wavefront into.
    """

    def __init__(self, n_modes=None, n_zernike_modes=None):
        if n_modes is None and n_zernike_modes is None:
            raise ValueError("either n_modes or n_zernike_modes must be specified")
        if n_zernike_modes is not None:
            n_modes = n_zernike_modes
        super().__init__(n_modes)

    def generator(self, nModes, circular_mask, user_mask, **kwargs):
        return ZernikeGenerator(circular_mask)

    def _numpy2coefficients(self, coeff_array):
        return ZernikeCoefficients(coeff_array)






import glob
from astropy.io import fits as pyfits
import logging


class SnapshotTweaker():

    def __init__(self,
                 pathname,
                 logger=logging.getLogger('SnapshotTweaker')):
        self._pathname = pathname
        self._logger = logger
        self._matchingFiles = glob.glob(self._pathname)

    def header_update_key(self, key, value):
        for filename in self._matchingFiles:
            self._logger.info('update %s: %s= %s' %
                              (filename, key, str(value)))
            pyfits.setval(filename, key, value)

    def header_delete_key(self, key):
        for filename in self._matchingFiles:
            self._logger.info('delete %s: %s' % (filename, key))
            pyfits.delval(filename, key)

    def header_rename_key_2(self, key, newkey):
        for filename in self._matchingFiles:
            self._logger.info('rename %s: %s= %s' % (filename, key, newkey))
            try:
                value = pyfits.getval(filename, key)
                pyfits.setval(filename, newkey, value)
                pyfits.delval(filename, key)
            except Exception:
                pass

    def header_rename_key(self, oldKey, newKey):
        for filename in self._matchingFiles:
            self._logger.info('rename %s: %s= %s' %
                              (filename, oldKey, newKey))
            try:
                hdulist = pyfits.open(filename, 'update')
                hdr = hdulist[0].header
                original_hdr = hdr.copy()
                for k, v in original_hdr.items():
                    newK = k.replace(oldKey, newKey)
                    if newK in hdr:
                        self._logger.debug(
                            '%s already exists. Skipped' % newK)
                        continue
                    if newK != k:
                        self._logger.debug(
                            'Replacing %s with %s. '
                            'New entry is: %s = %s' % (
                                oldKey, newKey, newK, v))
                        del hdr[k]
                        hdr['HIERARCH ' + newK] = v
                hdulist.close(output_verify='fix')
            except Exception as e:
                #self._logger.warning('headerRename failed (%s)') % e
                self._logger.warning(e)
                pass

    def header_get_key(self, key):
        ret = {}
        for filename in self._matchingFiles:
            ret[filename] = pyfits.getval(filename, key)
        return ret



import numpy as np
import astropy.units as u
from arte.dataelab.base_timeseries import BaseTimeSeries

class BaseIntensities(BaseTimeSeries):
    '''
    Time series for subaperture intensities
    '''
    def __init__(self, data, time_vector=None, astropy_unit=u.adu, data_label='Subaperture intensities', axes=None):
        super().__init__(data=data,
                         time_vector=time_vector,
                         astropy_unit=astropy_unit,
                         data_label=data_label,
                         axes=axes)

    def total_adu(self, threshold=0.1):
        '''
        Total number of ADU/frame. Optional threshold is relative to max
        '''
        data = self.time_average()
        data[np.where(data < data.max()*threshold)] = 0
        return data.sum()




import os
import astropy.units as u
from arte.photometry.transmissive_elements import Bandpass, TransmissiveElement, \
    TransmissiveSystem, Direction
from arte.utils.package_data import dataRootDir
from synphot.spectrum import SpectralElement
from arte.photometry.transmittance_calculator import interface_glass_to_glass, \
    internal_transmittance_calculator
from synphot.models import Empirical1D
from arte.photometry.filters import Filters


class RestoreTransmissiveElements(object):

    @classmethod
    def transmissive_elements_folder(cls):
        rootDir = dataRootDir()
        dirname = os.path.join(rootDir, 'photometry', 'transmissive_elements')
        return dirname

    @classmethod
    def restore_transmissive_elements_from_fits(cls, foldername):
        filename = os.path.join(foldername, '0.fits')
        return TransmissiveElement.from_fits(filename)

    @classmethod
    def restore_transmittance_from_dat(cls, foldername, wavelength_unit):
        filename = os.path.join(foldername, 't.dat')
        return SpectralElement.from_file(filename, wave_unit=wavelength_unit)

    @classmethod
    def restore_reflectance_from_dat(cls, foldername, wavelength_unit):
        filename = os.path.join(foldername, 'r.dat')
        return SpectralElement.from_file(filename, wave_unit=wavelength_unit)

    @classmethod
    def restore_refractive_index_from_dat(cls, foldername, wavelength_unit):
        filename = os.path.join(foldername, 'n.dat')
        return SpectralElement.from_file(filename, wave_unit=wavelength_unit)    


class CoatingsCatalog():
    
    @classmethod
    def _CoatingsFolder(cls, foldername):
        return os.path.join(
            RestoreTransmissiveElements.transmissive_elements_folder(),
            'coatings',
            foldername)
        
    @classmethod
    def ar_coating_589nm_001(cls):
        '''
        Narrowband (589 nm) AR coating. This is a simplified version, i.e.
        a peak at 589 nm.
        Data from Cedric's spreadsheet "background_calc_maory_v12.xls".
        '''
        t = RestoreTransmissiveElements.restore_transmittance_from_dat(
            cls._CoatingsFolder('ar_589nm_001'), u.um)
        a = Bandpass.zero()
        te = TransmissiveElement(transmittance=t, absorptance=a)
        return te
    
    @classmethod
    def ar_coating_589nm_002(cls):
        '''
        Narrowband (589 nm) AR coating. This is a simplified version, i.e.
        a peak of 0.995 (as indicated in E-MAO-SF0-INA-DER-001_02 MAORY  System
        Optical Design and Analysis Report) at 589 nm.
        '''
        t = Bandpass.top_hat(589 * u.nm, 1 * u.nm, 0.995, 0)
        a = Bandpass.zero()
        te = TransmissiveElement(absorptance=a, transmittance=t)
        return te

    @classmethod
    def ar_coating_broadband_001(cls):
        '''
        Broadband AR coating for CPM.
        Data from Demetrio, received by email on 12/04/2023.
        '''
        t = RestoreTransmissiveElements.restore_transmittance_from_dat(
            cls._CoatingsFolder('ar_broadband_001'), u.um)
        a = Bandpass.zero()
        te = TransmissiveElement(transmittance=t, absorptance=a)
        return te

    @classmethod
    def ar_coating_swir_001(cls):
        '''
        SWIR AR coating.
        Data from Edmund Optics Website.
        '''
        r = RestoreTransmissiveElements.restore_reflectance_from_dat(
            cls._CoatingsFolder('ar_swir_001'), u.um)
        a = Bandpass.zero()
        te = TransmissiveElement(reflectance=r, absorptance=a)
        return te

    @classmethod
    def ar_coating_nir_i_001(cls):
        '''
        NIR I AR coating.
        Data from Edmund Optics Website.
        '''
        r = RestoreTransmissiveElements.restore_reflectance_from_dat(
            cls._CoatingsFolder('ar_nir_i_001'), u.um)
        a = Bandpass.zero()
        te = TransmissiveElement(reflectance=r, absorptance=a)
        return te


class CoatedGlassesCatalog():
    
    @classmethod
    def _CoatedGlassesFolder(cls, foldername):
        return os.path.join(
            RestoreTransmissiveElements.transmissive_elements_folder(),
            'coated_glasses',
            foldername)
        
    @classmethod
    def infrasil_1mm_B_coated_001(cls):
        '''
        Infrasil window B-coated.
        Thickness: 1 mm.
        
        Data from Thorlabs website.
        '''
        t = RestoreTransmissiveElements.restore_transmittance_from_dat(
            cls._CoatedGlassesFolder('infrasil_1mm_B_coated_001'), u.um)
        r = Bandpass.zero()
        te = TransmissiveElement(transmittance=t, reflectance=r)
        return te
    
    @classmethod
    def infrasil_1mm_C_coated_001(cls):
        '''
        Infrasil window C-coated.
        Thickness: 1 mm.
        
        Data from Thorlabs website.
        '''
        t = RestoreTransmissiveElements.restore_transmittance_from_dat(
            cls._CoatedGlassesFolder('infrasil_1mm_C_coated_001'), u.um)
        r = Bandpass.zero()
        te = TransmissiveElement(transmittance=t, reflectance=r)
        return te
        

class GlassesCatalog():
    
    @classmethod
    def _GlassesFolder(cls, foldername):
        return os.path.join(
            RestoreTransmissiveElements.transmissive_elements_folder(),
            'glasses',
            foldername)
        
    @classmethod
    def cdgm_HQK3L_7mm_internal_001(cls):
        '''
        CDGM H-QK3L substrate of 7 mm thickness.
        Transmittance is internal.
        Data from RefractiveIndex website.
        '''
        t = RestoreTransmissiveElements.restore_transmittance_from_dat(
            cls._GlassesFolder('cdgm_HQK3L_7mm_internal_001'), u.um)
        r = Bandpass.zero()
        te = TransmissiveElement(transmittance=t, reflectance=r)
        return te    

    @classmethod
    def ohara_SFTM16_3mm_internal_001(cls):
        '''
        Ohara S-FTM16 substrate of 3 mm thickness.
        Transmittance is internal.
        Data from RefractiveIndex website.
        '''
        t = RestoreTransmissiveElements.restore_transmittance_from_dat(
            cls._GlassesFolder('ohara_SFTM16_3mm_internal_001'), u.um)
        r = Bandpass.zero()
        te = TransmissiveElement(transmittance=t, reflectance=r)
        return te    
        
    @classmethod
    def infrasil_1mm_001(cls):
        '''
        Infrasil window.
        Thickness: 1 mm.
        Transmittance is external.
        
        Data from Thorlabs website.
        '''
        t = RestoreTransmissiveElements.restore_transmittance_from_dat(
            cls._GlassesFolder('infrasil_1mm_001'), u.um)
        a = Bandpass.zero()
        te = TransmissiveElement(transmittance=t, absorptance=a)
        return te

    @classmethod
    def suprasil3001_3mm_internal_001(cls):
        '''
        Suprasil 3001 substrate of 3 mm thickness.
        Transmittance is internal.
        Data extrapolated from 10 mm curves.
        '''
        supra10mm = cls.suprasil3002_10mm_internal_001()
        wv = supra10mm.waveset
        t1 = supra10mm.transmittance(wv)
        l1 = 10
        l2 = 3
        t2 = internal_transmittance_calculator(l1, l2, t1)
        t = SpectralElement(
            Empirical1D, points=wv,
            lookup_table=t2)
        r = Bandpass.zero()
        te = TransmissiveElement(reflectance=r, transmittance=t)
        return te
        
    @classmethod
    def suprasil3002_10mm_internal_001(cls):
        '''
        Suprasil 3002 substrate of 10 mm thickness.
        Transmittance is internal.
        Data from Heraeus website.
        '''
        t = RestoreTransmissiveElements.restore_transmittance_from_dat(
            cls._GlassesFolder('suprasil3002_10mm_internal_001'), u.um)
        r = Bandpass.zero()
        te = TransmissiveElement(transmittance=t, reflectance=r)
        return te

    @classmethod
    def suprasil3002_40mm_internal_001(cls):
        '''
        Suprasil 3002 substrate of 40 mm thickness.
        Transmittance is internal.
        Data from Heraeus website. 
        '''
        t = RestoreTransmissiveElements.restore_transmittance_from_dat(
            cls._GlassesFolder('suprasil3002_40mm_internal_001'), u.um)
        r = Bandpass.zero()
        te = TransmissiveElement(transmittance=t, reflectance=r)
        return te

    @classmethod
    def suprasil3002_60mm_internal_001(cls):
        '''
        Suprasil 3002 substrate of 60 mm thickness.
        Transmittance is internal.
        Data from Heraeus website. 
        '''
        t = RestoreTransmissiveElements.restore_transmittance_from_dat(
            cls._GlassesFolder('suprasil3002_60mm_internal_001'), u.um)
        r = Bandpass.zero()
        te = TransmissiveElement(transmittance=t, reflectance=r)
        return te
    
    @classmethod
    def suprasil3002_70mm_internal_001(cls):
        '''
        Suprasil 3002 substrate of 70 mm thickness.
        Transmittance is internal.
        Data from Heraeus website. 
        '''
        t = RestoreTransmissiveElements.restore_transmittance_from_dat(
            cls._GlassesFolder('suprasil3002_70mm_internal_001'), u.um)
        r = Bandpass.zero()
        te = TransmissiveElement(transmittance=t, reflectance=r)
        return te

    @classmethod
    def suprasil3002_80mm_internal_001(cls):
        '''
        Suprasil 3002 substrate of 80 mm thickness.
        Transmittance is internal.
        '''
        t = RestoreTransmissiveElements.restore_transmittance_from_dat(
            cls._GlassesFolder('suprasil3002_80mm_internal_001'), u.um)
        r = Bandpass.zero()
        te = TransmissiveElement(transmittance=t, reflectance=r)
        return te
    
    @classmethod
    def suprasil3002_85mm_internal_001(cls):
        '''
        Suprasil 3002 substrate of 85 mm thickness.
        Transmittance is internal.
        Data has been extrapolated from 10 mm Suprasil 3002 data that has been
        collected from Heraeus website.
        '''
        t = RestoreTransmissiveElements.restore_transmittance_from_dat(
            cls._GlassesFolder('suprasil3002_85mm_internal_001'), u.um)
        r = Bandpass.zero()
        te = TransmissiveElement(transmittance=t, reflectance=r)
        return te
        
    @classmethod
    def suprasil3002_108mm_internal_001(cls):
        '''
        Suprasil 3002 substrate of 108 mm thickness.
        Transmittance is internal.
        Data from Heraeus website. 
        '''
        t = RestoreTransmissiveElements.restore_transmittance_from_dat(
            cls._GlassesFolder('suprasil3002_108mm_internal_001'), u.um)
        r = Bandpass.zero()
        te = TransmissiveElement(transmittance=t, reflectance=r)
        return te
    
    @classmethod
    def ohara_quartz_SK1300_10mm_internal_001(cls):
        '''
        Ohara quartz SK-1300 substrate of 10 mm thickness.
        Transmittance is internal.
        Data from Ohara website. 
        '''
        t = RestoreTransmissiveElements.restore_transmittance_from_dat(
            cls._GlassesFolder('ohara_quartz_SK1300_10mm_internal_001'), u.um)
        r = Bandpass.zero()
        te = TransmissiveElement(transmittance=t, reflectance=r)
        return te

    @classmethod
    def ohara_quartz_SK1300_85mm_internal_001(cls):
        '''
        Ohara quartz SK-1300 substrate of 85 mm thickness.
        Transmittance is internal.
        Data derived from transmittance data for 10 mm thickness. 
        '''
        t = RestoreTransmissiveElements.restore_transmittance_from_dat(
            cls._GlassesFolder('ohara_quartz_SK1300_85mm_internal_001'), u.um)
        r = Bandpass.zero()
        te = TransmissiveElement(transmittance=t, reflectance=r)
        return te

    @classmethod
    def ohara_SFPL51_10mm_internal_001(cls):
        '''
        Ohara SFPL-51 substrate of 10 mm thickness.
        Transmittance is internal.
        Data from RefractiveInfo website. 
        '''
        t = RestoreTransmissiveElements.restore_transmittance_from_dat(
            cls._GlassesFolder('ohara_SFPL51_10mm_internal_001'), u.um)
        r = Bandpass.zero()
        te = TransmissiveElement(transmittance=t, reflectance=r)
        return te
    
    @classmethod
    def ohara_PBL35Y_10mm_internal_001(cls):
        '''
        Ohara PBL-35Y substrate of 10 mm thickness.
        Transmittance is internal.
        Data from RefractiveInfo website. 
        '''
        t = RestoreTransmissiveElements.restore_transmittance_from_dat(
            cls._GlassesFolder('ohara_PBL35Y_10mm_internal_001'), u.um)
        r = Bandpass.zero()
        te = TransmissiveElement(transmittance=t, reflectance=r)
        return te
    
    @classmethod
    def ohara_PBL35Y_3mm_internal_001(cls):
        '''
        Ohara PBL-35Y substrate of 3 mm thickness.
        Transmittance is internal.
        Data from RefractiveInfo website. 
        '''
        t = RestoreTransmissiveElements.restore_transmittance_from_dat(
            cls._GlassesFolder('ohara_PBL35Y_3mm_internal_001'), u.um)
        r = Bandpass.zero()
        te = TransmissiveElement(transmittance=t, reflectance=r)
        return te
    
    @classmethod
    def schott_NSF2_9dot8_mm_internal_001(cls):
        '''
        Schott N-SF2 substrate of 9.8 mm thickness.
        Transmittance is internal.
        Data from RefractiveInfo website. 
        '''
        t = RestoreTransmissiveElements.restore_transmittance_from_dat(
            cls._GlassesFolder('schott_NSF2_9.8mm_internal_001'), u.um)
        r = Bandpass.zero()
        te = TransmissiveElement(transmittance=t, reflectance=r)
        return te
    
    @classmethod
    def schott_NPSK53A_10_mm_internal_001(cls):
        '''
        Schott N-PSK53A substrate of 10 mm thickness.
        Transmittance is internal.
        Data from RefractiveInfo website. 
        '''
        t = RestoreTransmissiveElements.restore_transmittance_from_dat(
            cls._GlassesFolder('schott_NPSK53A_10mm_internal_001'), u.um)
        r = Bandpass.zero()
        te = TransmissiveElement(transmittance=t, reflectance=r)
        return te

    @classmethod
    def interface_ohara_SFPL51_to_ohara_PBL35Y_001(cls):
        '''
        Interface between Ohara SFPL-51 and Ohara PBL-35Y.
        Data from RefractiveInfo website. 
        '''
        n_sfpl51 = RestoreTransmissiveElements.restore_refractive_index_from_dat(
            cls._GlassesFolder('ohara_SFPL51_10mm_internal_001'), u.um)
        n_pbl35y = RestoreTransmissiveElements.restore_refractive_index_from_dat(
            cls._GlassesFolder('ohara_PBL35Y_10mm_internal_001'), u.um)
        wv = n_sfpl51.waveset
        r = SpectralElement(
            Empirical1D, points=wv,
            lookup_table=interface_glass_to_glass(n_sfpl51(wv), n_pbl35y(wv)))
        a = Bandpass.zero()
        te = TransmissiveElement(reflectance=r, absorptance=a)
        return te
    
    @classmethod
    def interface_schott_NSF2_to_schott_NPSK53A_001(cls):
        '''
        Interface between Schott N-SF2 and Schott N-PSK53A.
        Data from RefractiveInfo website. 
        '''
        n_sf2 = RestoreTransmissiveElements.restore_refractive_index_from_dat(
            cls._GlassesFolder('schott_NSF2_9.8mm_internal_001'), u.um)
        n_npsk53a = RestoreTransmissiveElements.restore_refractive_index_from_dat(
            cls._GlassesFolder('schott_NPSK53A_10mm_internal_001'), u.um)
        wv = n_sf2.waveset
        r = SpectralElement(
            Empirical1D, points=wv,
            lookup_table=interface_glass_to_glass(n_sf2(wv), n_npsk53a(wv)))
        a = Bandpass.zero()
        te = TransmissiveElement(reflectance=r, absorptance=a)
        return te


class FiltersCatalog():
    
    @classmethod
    def bessel_H(cls):
        t = Filters.get(Filters.BESSEL_H)
        a = Bandpass.zero()
        te = TransmissiveElement(transmittance=t, absorptance=a)
        return te




import numpy as np
import astropy.units as u
from arte.dataelab.base_timeseries import BaseTimeSeries


def _mapper2d_square(pixels):
    npix = np.sqrt(len(pixels))
    if npix != int(npix):
        raise ValueError('Cannot remap pixels into 2d, frame is not square')
    npix = int(npix)
    return pixels.reshape(npix, npix)


class BasePixels(BaseTimeSeries):
    '''
    Time series for pixel data
    '''
    def __init__(self, data, time_vector=None, astropy_unit=u.adu, data_label='Pixel values'):

        super().__init__(data=data,
                         time_vector=time_vector,
                         astropy_unit=astropy_unit,
                         data_label=data_label)

    def total_adu(self, threshold=0.1):
        '''Total number of ADU/frame. Optional threshold is relative to max.'''
        data = self.time_average()
        data[np.where(data < data.max()*threshold)] = 0
        return data.sum()

# __oOo__



import os
import abc
import zipfile
from pathlib import Path

import numpy as np
from astropy.io import fits
from arte.utils.help import add_help
from arte.dataelab.dataelab_utils import is_dataelab
from arte.utils.not_available import NotAvailable


@add_help
class DataLoader():
    '''
    Abstract base class for data loaders
    '''
    def __init__(self):
        pass

    @abc.abstractmethod
    def assert_exists(self):
        '''Assert that the data is available'''

    @abc.abstractmethod
    def filename(self):
        '''Return the data filename, if available'''

    @abc.abstractmethod
    def load(self):
        '''Load data and return it'''


class FitsDataLoader(DataLoader):
    '''Loader for data stored into FITS files

    Parameters
    ----------

    filename: str
        FITS filename or full path
    ext: int, optional
        FITS extenstion to read. Defaults to the primary card
    transpose_axes: tuple, optional
        Axes transpose pattern as specified for np.transpose.
        Use this if time is not your first dimension
    postprocess: function, optional
        function to call after loading data and before returning it
        Must take a single parameter with the whole data array
        and return the processed data array.
    '''
    def __init__(self, filename, ext=None, transpose_axes=None, postprocess=None):
        super().__init__()
        if isinstance(filename, Path):
            self._filename = str(filename)
        else:
            self._filename = filename
        self._ext = ext
        self._transpose_axes = transpose_axes
        self._postprocess = postprocess

    def assert_exists(self):
        assert os.path.exists(self._filename), 'File Not Found: ' + self._filename

    def filename(self):
        return self._filename

    def load(self):
        if self._ext:
            data = fits.getdata(self._filename, ext=self._ext)
        else:
            data = fits.getdata(self._filename)
        if self._transpose_axes is not None:
            print(data.shape, self._transpose_axes)
            data = data.transpose(*self._transpose_axes)
        if self._postprocess:
            data = self._postprocess(data)
        return data


class NumpyDataLoader(DataLoader):
    '''Loader for data stored into np or npz files

    Parameters
    ----------
    filename: str
        numpy filename or full path
    key: str, optional
        array key for .npz files
    transpose_axes: tuple, optional
        Axes transpose pattern as specified for np.transpose.
        Use this if time is not your first dimension
    postprocess: function, optional
        function to call after loading data and before returning it
        Must take a single parameter with the whole data array
        and return the processed data array.
    '''
    def __init__(self, filename, key=None, transpose_axes=None, postprocess=None):
        super().__init__()
        if isinstance(filename, Path):
            self._filename = str(filename)
        else:
            self._filename = filename
        if self._filename.endswith('.npz') and key is None:
            key = 'arr_0'
        self._key = key
        self._transpose_axes = transpose_axes
        self._postprocess = postprocess

    def assert_exists(self):
        assert os.path.exists(self._filename), 'File Not Found: ' + self._filename
        if self._filename.endswith('.npz'):
            npyname = self._key+'.npy'
            assert npyname in zipfile.ZipFile(self._filename).namelist(), \
                 f'Key not found: {npyname} in file: {self._filename}'

    def filename(self):
        return self._filename

    def load(self):
        if self._key:
            data = np.load(self._filename)[self._key]
        else:
            data = np.load(self._filename)
        if self._transpose_axes is not None:
            data = data.transpose(*self._transpose_axes)
        if self._postprocess:
            data = self._postprocess(data)
        return data

        
class TxtDataLoader(DataLoader):
    '''Loader for data stored into txt files

    Parameters
    ----------
    filename: str
        numpy filename or full path
    transpose_axes: tuple, optional
        Axes transpose pattern as specified for np.transpose.
        Use this if time is not your first dimension
    postprocess: function, optional
        function to call after loading data and before returning it
        Must take a single parameter with the whole data array
        and return the processed data array.
    '''
    def __init__(self, filename, transpose_axes=None, postprocess=None):
        super().__init__()
        if isinstance(filename, Path):
            self._filename = str(filename)
        else:
            self._filename = filename
        self._transpose_axes = transpose_axes
        self._postprocess = postprocess

    def assert_exists(self):
        assert os.path.exists(self._filename), 'File Not Found: ' + self._filename

    def filename(self):
        return self._filename

    def load(self):
        data = np.loadtxt(self._filename)
        if self._transpose_axes is not None:
            data = data.transpose(*self._transpose_axes)
        if self._postprocess:
            data = self._postprocess(data)
        return data


class DummyLoader(DataLoader):
    '''Dummy loader for data not stored anywhere'''
    def __init__(self):
        super().__init__()

    def assert_exists(self):
        pass

    def filename(self):
        return None

    def load(self):
        return None

class OnTheFlyLoader(DataLoader):
    '''Loader for data calculated on the fly'''
    def __init__(self, func):
        super().__init__()
        self._func = func

    def assert_exists(self):
        pass

    def filename(self):
        return None

    def load(self):
        return self._func()


class ConstantDataLoader(DataLoader):
    '''Loader for constant data'''
    def __init__(self, data):
        super().__init__()
        self._data = data

    def assert_exists(self):
        pass

    def filename(self):
        return None

    def load(self):
        return self._data


def data_loader_factory(obj, allow_none=False, name=''):
    '''
    Return the correct DataLoader instance for *obj*, which might be:
    * a string with a filename among the known ones (fits or npy)
    * a pathlib.Path instance
    * a numpy array
    * a DataLoader instance (returned unchanged)
     
    Raises ValueError if the guess fails,

    Parameters
    ----------
    obj: Loader class, numpy array, str (filename), Path instance, or None
        object to wrap in a DataLoader class
    allow_none: bool, optional
        if set to True, obj can be None. If this flag is False and obj
        is None, a ValueError exception willb be raised.
    name: str, optional
        object name for error messages.
        
    Returns
    -------
    DataLoader instance
        
    '''
    if isinstance(obj, Path):
        obj = str(obj)
    if isinstance(obj, str):
        if obj.endswith('.fits'):
            return FitsDataLoader(obj)
        elif obj.endswith('.npy') or obj.endswith('.npz'):
            return NumpyDataLoader(obj)
        elif obj.endswith('.txt'):
            return TxtDataLoader(obj)
        else:
            raise ValueError(f'Cannot guess correct DataLoader instance for filename {obj}')
    elif isinstance(obj, np.ndarray):
        return ConstantDataLoader(obj)
    elif isinstance(obj, DataLoader):
        return obj
    elif callable(obj):
        return OnTheFlyLoader(obj)
    elif is_dataelab(obj):
        return OnTheFlyLoader(obj.get_data)
    elif obj is None and allow_none is True:
        return obj
    else:
        name = name or 'object'
        if allow_none:
            errstr = f'{name} must be a Loader class, a numpy array, or a filename or Path instance, or None.'
        else:
            errstr = f'{name} must be a Loader class, a numpy array, or a filename or Path instance.'
        raise ValueError(errstr)

def data_axes(obj):
    '''
    Returns
    -------
    axes: tuple(str) or None
        axes names tuple
    '''
    if is_dataelab(obj):
        return obj.axes
    else:
        return None

# __oOo__



import numbers
from functools import cached_property, cache
import numpy as np
from astropy import units as u

from arte.time_series.time_series import TimeSeries
from arte.utils.help import modify_help
from arte.utils.not_available import NotAvailable
from arte.dataelab.data_loader import data_loader_factory, data_axes
from arte.dataelab.unit_handler import UnitHandler
from arte.dataelab.dataelab_utils import is_dataelab
from arte.time_series.indexer import DefaultIndexer
from arte.utils.displays import movie, tile, savegif
from arte.utils.show_array import show_array

class BaseTimeSeries(TimeSeries):
    '''
    Generic time series.

    A time series holds:
        - N samples of numeric data of arbitrary type and dimensions, with an optional astropy unit.
        - sampling time for each sample
        - optionally, a data label for plots and displays

    Data is accessed with the get_data() method. Derived classes can define specialized
    arguments to return data subsets, e.g. get_data(quadrant=2)
    Basic arithmetic operations (sum, div, etc) are supported. Astropy units, if present,
    will be enforced.

    Parameters
    ----------
    data_loader: instance of DataLoader or derived class, or numpy array, or filename (string or pathlib instance)
         time series data [time, data_d0 [, data_d1...]]
    time_vector: instance of DataLoader or derived class, or numpy_array, or None
         time vector data. If None, a default counter from 0 to N-1 samples will be assigned
    astropy_unit: astropy unit or None
         if possible, astropy unit to use with the data.
    data_label: string or None
         human-readable label for plot (e.g.: "Surface modal coefficients" )
    axes: sequence or None
         sequence of axes names, optional
    '''
    def __init__(self, data, time_vector=None, astropy_unit=None, data_label=None, axes=None):

        data_loader = data_loader_factory(data, allow_none=False, name='data')
        time_vector_loader = data_loader_factory(time_vector, allow_none=True, name='time_vector')
        if axes is None:
            # Try to derive from input data
            axes = data_axes(data)

        try:
            super().__init__(axes=axes)

            # Also test that the data file is there, when possible
            _ = data_loader.assert_exists()
            if time_vector_loader is not None:
                _ = time_vector_loader.assert_exists()

        except AssertionError as e:
            print(e)
            NotAvailable.transformInNotAvailable(self)
            return

        self._data_loader = data_loader
        self._time_vector_loader = time_vector_loader
        self._astropy_unit = astropy_unit
        self._data_label = data_label
        self._unit_handler = UnitHandler(wanted_unit = astropy_unit)
        self._default_indexer = DefaultIndexer()

    def filename(self):
        '''Data filename (full path)'''
        return self._data_loader.filename()

    @cached_property
    def shape(self):
        '''Data series shape

        TODO depending on the data loader, it could be found
        without loading the entire data array
        '''
        return self.get_data().shape

    @cache
    def _get_not_indexed_data(self):
        '''Lazy loading of raw data and astropy unit application'''
        return self._unit_handler.apply_unit(self._data_loader.load())

    def _get_time_vector(self):
        '''Lazy loading of time vector'''
        if self._time_vector_loader is not None:
            return self._time_vector_loader.load()
        else:
            return super()._get_time_vector()

    def get_index_of(self, *args, **kwargs):
        '''Return a selection index

        Parameters
        ----------
        *args: tuple
           user-defined arguments for data selection
        **kwargs: dict, optional
           extra user-defined arguments for data selection

        Returns
        -------
        index: None, integer, list of integers, slice objects, or a tuple of the previous possibilities.
           indexes to select a data subset. If None, all data is selected.
           An integer, a list of integers or a slice will select those rows.
           If the data is multidimensional, a tuple can be returned where each element
           will select data across a single dimension.
           For a detailed explanation, see the "Advanced indexing" topic at
           https://numpy.org/doc/stable/user/basics.indexing.html#advanced-indexing
        '''
        return self._default_indexer.elements(*args, **kwargs)

    def data_label(self):
        return self._data_label

    def data_unit(self):
        '''Data unit string (for plots)'''
        return self._unit_handler.actual_unit_name()

    def astropy_unit(self):
        '''Data unit as an astropy unit'''
        return self._unit_handler.actual_unit()

    # Override to provide custom displays
    def _get_display_cube(self, data_to_display):
        '''Generate a 3d cube for display'''
        return np.atleast_3d(data_to_display)

    @modify_help(arg_str='[series_idx], [times=[from, to]]')
    def get_display(self, *args, times=None, **kwargs):
        '''Display cube for the specified time interval'''
        data_to_display = self.get_data(*args, times=times, **kwargs)
        display_data = self._get_display_cube(data_to_display)
        if isinstance(display_data, u.Quantity):
            return display_data.value
        else:
            return display_data
    
    def get_display_axes(self):
        '''Display cube axes names as a list of 3 strings'''
        return ('time', '', '')

    @modify_help(arg_str='[series_idx], [times=[from, to]]')
    def movie(self, *args, interval=0.1, **kwargs):
        '''Display data as a movie'''
        frames = self.get_display(*args, **kwargs)
        movie(frames, interval=interval)

    @modify_help(arg_str='[series_idx], [times=[from, to]]')
    def tile(self, *args, rowlength=10, **kwargs):
        '''Display data as a tiled 2d frame'''
        frames = self.get_display(*args, **kwargs)
        return tile(frames, rowlength=rowlength)

    @modify_help(arg_str='filename, [series_idx], [times=[from, to]]')
    def savegif(self, filename, *args, interval=0.1, loop=0, **kwargs):
        '''Save data as an animated GIF'''
        frames = self.get_display(*args, **kwargs)
        savegif(frames, filename, interval=interval, loop=loop)

    def _check_broadcast(self, other):
        _ = np.broadcast_shapes(self.shape, other.shape)

    def _check_mult(self, other):
        if not self.shape[1] == other.shape[0]:
            raise ValueError(f'Data dimensions do not match: {self.shape} and {other.shape}')

    def _check_none(self, other):
        return None

    def _check_and_execute(self, check, other, func):
        check(other)
        return func(self.get_data(), other.get_data())

    def _operator(self, other, func, check):
        if isinstance(other, (numbers.Number, u.Quantity, np.ndarray)):
            new_data = func(self.get_data(), other)
        elif not is_dataelab(other):
            return NotImplemented
        else:
            new_data = lambda: self._check_and_execute(check, other, func)

        if self.__class__ == other.__class__:
            new_class = self.__class__
        else:
            new_class = BaseTimeSeries
        return new_class(new_data, time_vector=self._get_time_vector,
                         astropy_unit=None,
                         data_label=self._data_label)

    def __add__(self, other):
        return self._operator(other, lambda x, y: x + y, self._check_broadcast)

    def __sub__(self, other):
        return self._operator(other, lambda x, y: x - y, self._check_broadcast)

    def __mul__(self, other):
        return self._operator(other, lambda x, y: x * y, self._check_broadcast)

    def __truediv__(self, other):
        return self._operator(other, lambda x, y: x / y, self._check_broadcast)

    def __floordiv__(self, other):
        return self._operator(other, lambda x, y: x // y, self._check_broadcast)

    def __mod__(self, other):
        return self._operator(other, lambda x, y: x % y, self._check_broadcast)

    def __matmul__(self, other):
        return self._operator(other, lambda x, y: x @ y, self._check_mult)

    def __pow__(self, other):
        return self._operator(other, lambda x, y: x ** y, self._check_broadcast)

    def __neg__(self):
        return self._operator(self, lambda x, y: -x, self._check_none)

    def __abs__(self):
        return self._operator(self, lambda x, y: abs(x), self._check_none)

    def imshow(self, *args, cut_wings=0, title='', xlabel='', ylabel='', **kwargs):
        '''
        Display X and Y slope 2d images
        cut_wings=x means that colorbar is saturated for array values below x percentile
        and above 100-x percentile. Default is 0, i.e. all data are displayed; values below
        0 are forced to 0, values above 50 are set to 50.
        '''
        array2show = self.get_display(*args, **kwargs).mean(axis=0)
        _, default_xlabel, default_ylabel = self.get_display_axes()
        if not xlabel:
            xlabel = default_xlabel
        if not ylabel:
            ylabel = default_ylabel
        if not title:
            title = self.data_label()

        return show_array(array2show, cut_wings, title, xlabel, ylabel, self.data_unit())

    @modify_help(call='plot_hist([series_idx], from_freq=xx, to_freq=xx, )')
    def plot_hist(self, *args, from_t=None, to_t=None,
                  overplot=None, plot_to=None,
                  label=None,  **kwargs):
        '''Plot histogram. TODO: does not work, rewrite'''
        hist = self.get_data(*args, **kwargs)
        t = self._get_time_vector()
        if from_t is None: from_t=t.min()
        if to_t is None: to_t=t.max()
        ul = t <= to_t
        dl = t >= from_t
        lim = ul & dl
        t = t[lim]
        hist = hist[lim]

        from matplotlib.axes import Axes
        if plot_to is None:
            import matplotlib.pyplot as plt
        else:
            plt = plot_to
        if not overplot:
            plt.cla()
            plt.clf()

        lines = plt.plot(t, hist)
        if self.data_unit() is None:
            units = "(a.u.)"
        else:
            units = self.data_unit()

        title = 'Time history'
        xlabel = 'time [s]'
        ylabel = '['+units+']'
        if self.data_label() is not None:
            title = title + " of " + self.data_label()
            ylabel = self.data_label() + " " + ylabel

        if isinstance(plt, Axes):
            plt.set(title=title, xlabel=xlabel, ylabel=ylabel)
        else:
            plt.title(title)
            plt.xlabel(xlabel)
            plt.ylabel(ylabel)
            if label is not None:
                if isinstance(label, str):
                    plt.legend([label] * len(lines))
                else:
                    plt.legend(label)
        return plt

    @modify_help(call='plot_spectra([series_idx], from_freq=xx, to_freq=xx)')
    def plot_spectra(self, *args, from_freq=None, to_freq=None,
                     segment_factor=None,
                     overplot=False,
                     label=None, plot_to=None,
                     lineary=False, linearx=False,
                     **kwargs):
        '''Plot PSD'''
        power = self.power(from_freq, to_freq,
                           segment_factor,
                           *args, **kwargs)
        freq = self.last_cut_frequency()

        # linearx=True forces lineary to True
        if linearx: lineary=True
        if lineary and not linearx:
            p_shape = power.shape
            if len(p_shape)== 1:
                power *= freq
            else:
                power *= np.repeat(freq.reshape(p_shape[0],1), p_shape[1], axis=1)

        from matplotlib.axes import Axes

        if plot_to is None:
            import matplotlib.pyplot as plt
        else:
            plt = plot_to

        if not overplot:
            plt.cla()
            plt.clf()
        lines = plt.plot(freq[1:], power[1:])

        if lineary and not linearx:
            plt.semilogx()
        elif linearx and not lineary:
            plt.semilogy()
        elif not (linearx and lineary):
            plt.loglog()

        if self.data_unit() is None:
            units = "(a.u.)"
        else:
            units = self.data_unit()

        title = 'PSD'
        if self.data_label() is not None:
            title = title + " of " + self.data_label()
        xlabel = 'frequency [Hz]'
        if lineary and not linearx:
            ylabel = 'frequency*PSD ['+units+'^2]'
        else:
            ylabel = 'PSD ['+units+'^2]/Hz'
        if isinstance(plt, Axes):
            plt.set(title=title, xlabel=xlabel, ylabel=ylabel)
        else:
            plt.title(title)
            plt.xlabel(xlabel)
            plt.ylabel(ylabel)
            if label is not None:
                if isinstance(label, str):
                    plt.legend([label] * len(lines))
                else:
                    plt.legend(label)
        return plt

    @modify_help(call='plot_cumulative_spectra([series_idx], from_freq=xx, to_freq=xx)')
    def plot_cumulative_spectra(self, *args, from_freq=None, to_freq=None,
                                segment_factor=None,
                                label=None, plot_to=None,
                                overplot=False, plot_rms=False, lineary=False,
                                **kwargs):
        '''Plot cumulative PSD'''
        power = self.power(*args, from_freq=from_freq, to_freq=to_freq,
                           segment_factor=segment_factor,
                           **kwargs)
        freq = self.last_cut_frequency()
        freq_bin = freq[1]-freq[0] # frequency bin

        from matplotlib.axes import Axes

        if plot_to is None:
            import matplotlib.pyplot as plt
        else:
            plt = plot_to

        if not overplot:
            plt.cla()
            plt.clf()

        # cumulated PSD
        cumpsd = np.cumsum(power, 0) * freq_bin
        if plot_rms:
            # cumulated RMS
            cumpsd = np.sqrt(cumpsd)
            label_str = "RMS"
            label_pow = ""
        else:
            # cumulated PSD
            label_str = "Variance"
            label_pow = "^2"

        lines = plt.plot(freq[1:], cumpsd[1:])
        if lineary:
            plt.semilogx()
        else:
            plt.loglog()

        if self.data_unit() is None:
            units = "(a.u.)"
        else:
            units = self.data_unit()

        title = "Cumulated " + label_str
        if self.data_unit() is not None:
            title = title + " of " + self.data_label()
        xlabel = 'frequency [Hz]'
        ylabel = 'Cumulated '+label_str+' ['+units+label_pow+']'
        if isinstance(plt, Axes):
            plt.set(title=title, xlabel=xlabel, ylabel=ylabel)
        else:
            plt.title(title)
            plt.xlabel(xlabel)
            plt.ylabel(ylabel)
            if label is not None:
                if isinstance(label, str):
                    plt.legend([label] * len(lines))
                else:
                    plt.legend(label)
        return plt



# __oOo__



# -*- coding: utf-8 -*-

import inspect
from functools import wraps

import numpy as np
import astropy.units as u

from arte.utils.not_available import NotAvailable


def assert_unit_is_equivalent(var, ref):
    '''Make sure that `var` has a unit compatible with `ref`'''

    if isinstance(var, u.Quantity):
        if isinstance(ref, u.Quantity):
            assert var.unit.is_equivalent(ref.unit)
        else:
            raise AssertionError('Variable has units while it should not')
    else:
        if isinstance(ref, u.Quantity):
            raise AssertionError('Variable has no unit, while it should be '
                                 'equivalent to %s' % (ref.unit))


def separate_value_and_unit(var):
    '''Returns the argument value and unit, if `var` has one.

    If not, `var` is returned unchanged and unit is set to 1.

    Parameters
    ----------
    var: any type
        the variable to be tested

    Returns
    -------
    value, unit
        tuple with the value and the astropy unit of `var`
        or 1 if `var` does not have a unit.

    Examples
    --------
    > a = 42 * u.m
    > separate_value_and_unit(a)
    (42.0, Unit("m"))

    > b = 42
    > separate_value_and_unit(b)
    (42, 1)
    '''
    if isinstance(var, u.Quantity):
        return var.value, var.unit
    else:
        return var, 1


def make_sure_its_a(unit, v, name='', copy=True):
    '''
    Make sure that `v` has the astropy unit `unit`.

    If `v` does not have any unit, apply `unit` and return the combined value.
    If it has one, check that it can be converted to `unit`,
    and return the converted value. Otherwise, raise astropy.units.UnitsError.

    Parameters
    ----------
    unit: astropy unit
        the wanted unit
    v:
        the value under test
    name: string
        description of `v`. Will be used in error messages
        in case the conversion fails.

    Returns
    -------
    astropy Quantity
        the original value converted to `unit`.

    Raises
    ------
    astropy.units.UnitsError
        if the conversion to `unit` fails.

    '''
    # Special case for masked arrays, who wraps their data into a "data" attribute
    if isinstance(v, np.ma.MaskedArray):
        return np.ma.MaskedArray(
            make_sure_its_a(unit, v.data, name=name, copy=copy),
            mask=v.mask)

    if isinstance(v, NotAvailable):
        return v

    if not isinstance(v, u.Quantity):
        return u.Quantity(v, unit=unit, copy=copy)

    try:
        normalized = v.to(unit)
    except u.UnitConversionError:
        errmsg = 'Symbol {} should be {}, got {} instead'.format(
                  name, str(unit), str(v.unit))
        raise u.UnitsError(errmsg)
    return normalized


def unit_check(f):
    '''
    Decorator to add type checking of astropy units to a function.

    This decorator will ensure that, each time the decorated function is
    called, all the arguments have the correct astropy units,
    as defined in the default values of the
    decorated function, calling `make_sure_its_a()` for each of them.

    If the function does not define a unit for a parameter,
    the corresponding argument is not modified, whether it has a unit or not.

    If any of the checks fails, raises `TypeError` when the decorated function
    is called.
    '''
    # Keep as much as possible out of the wrapper, so it only executes
    # once at function definition time.
    sig = inspect.signature(f)
    pars_to_check = [p for p in sig.parameters.values()
                     if isinstance(p.default, u.Quantity)]

    @wraps(f)
    def wrapper(*args, **kwargs):

        # Reconstruct the decorated function's arguments
        bound = sig.bind(*args, **kwargs)
        bound.apply_defaults()

        # Check all arguments that have an astropy unit in their defaults
        for p in pars_to_check:
            bound.arguments[p.name] = make_sure_its_a(p.default.unit,
                                                      bound.arguments[p.name],
                                                      p.name)
        ret_value = f(*bound.args, **bound.kwargs)

        if sig.return_annotation is not inspect.Signature.empty:
            return make_sure_its_a(sig.return_annotation, ret_value,
                                   'Return value from function ' + f.__name__)
        else:
            return ret_value

    return wrapper



from arte.dataelab.base_timeseries import BaseTimeSeries
from arte.utils.not_available import NotAvailable
from arte.dataelab.cache_on_disk import cache_on_disk
from arte.dataelab.data_loader import OnTheFlyLoader

class BaseProjection(BaseTimeSeries):

    def __init__(self, source_timeseries, projection_matrix, astropy_unit=None, data_label=None):
        try:
            assert not isinstance(source_timeseries, NotAvailable)
            assert not isinstance(projection_matrix, NotAvailable)
            super().__init__(data=OnTheFlyLoader(self.project),
                             time_vector=OnTheFlyLoader(source_timeseries.get_time_vector),
                             astropy_unit=astropy_unit,
                             data_label=data_label)
            self._unit_handler.set_force(True)

        except AssertionError as e:
            NotAvailable.transformInNotAvailable(self)
            return
        self._source_timeseries = source_timeseries
        self._projection_matrix = projection_matrix

    @cache_on_disk
    def project(self):
        '''Perform projection'''
        data = self._source_timeseries.get_data()
        proj = self._projection_matrix.get_data()
        print(data, proj)
        if isinstance(data, NotAvailable) or isinstance(proj, NotAvailable):
            data = NotAvailable()
        else:
            data = data @ proj
        return data

# __oOo__



import numpy as np
import astropy.units as u

from arte.utils.help import add_help
from arte.utils.not_available import NotAvailable
from arte.dataelab.data_loader import data_loader_factory
from arte.dataelab.unit_handler import UnitHandler
from arte.utils.show_array import show_array
from arte.time_series.axis_handler import AxisHandler


@add_help
class BaseData():
    '''
    Generic static data

    parameters
    ----------
    data: instance of DataLoader or derived class, filename, or numpy array
        data to load
    astropy_unit: unit
        if possible, astropy unit to use with the data.
    data_label: str, optional
        human-readable label for plot (e.g.: "Surface modal coefficients")
    axes: sequence or None
         sequence of axes names, optional
    '''
    def __init__(self, data, astropy_unit=None, data_label=None, axes=None):
        
        data_loader = data_loader_factory(data, allow_none=False, name='data')

        try:
            # Test that the data file is there, when possible
            _ = data_loader.assert_exists()

        except AssertionError as e:
            print(e)
            NotAvailable.transformInNotAvailable(self)
            return
        
        if data_label is None:
            data_label = self.__class__.__name__

        self._data_loader = data_loader
        self._astropy_unit = astropy_unit
        self._data_label = data_label
        self._unit_handler = UnitHandler(wanted_unit = astropy_unit)
        self._axis_handler = AxisHandler(axes)

    def filename(self):
        '''Data filename (full path)'''
        return self._data_loader.filename()

    def get_data(self, *args, axes=None, **kwargs):
        '''Raw data, selecting elements based on the indexer'''

        raw_data = self._get_not_indexed_data()
        index = self.get_index_of(*args, **kwargs)
        if index is None:
            data = raw_data
        else:
            data = raw_data[index]
        return self._axis_handler.transpose(data, axes)

    def _get_not_indexed_data(self):
        return self._unit_handler.apply_unit(self._data_loader.load())

    def get_index_of(self, *args, **kwargs):
        '''Return data slice. Override in derived classes if needed'''
        return None

    @property
    def shape(self):
        return self.get_data().shape

    def data_label(self):
        '''Long-form data label (for plots)'''
        return self._data_label or ''

    def data_unit(self):
        '''Data unit string (for plots)'''
        return self._unit_handler.actual_unit_name() or ''

    def astropy_unit(self):
        '''Data unit as an astropy unit'''
        return self._unit_handler.actual_unit()

    # Override to provide custom displays
    def _get_display_frame(self, data_to_display):
        return np.atleast_2d(data_to_display)

    def get_display(self, *args, **kwargs):
        '''Data mapped in 2d'''
        data_to_display = self.get_data(*args, **kwargs)
        display_data = self._get_display_frame(data_to_display)
        if isinstance(display_data, u.Quantity):
            return display_data.value
        else:
            return display_data

    def imshow(self, cut_wings=0, title='', xlabel='', ylabel='', **kwargs):
        '''
        Display X and Y slope 2d images
        cut_wings=x means that colorbar is saturated for array values below x percentile
        and above 100-x percentile. Default is 0, i.e. all data are displayed; values below
        0 are forced to 0, values above 50 are set to 50.
        '''
        if not title:
            title = self.data_label()
        axes = self._axis_handler.axes()
        if not ylabel and axes and len(axes) > 0:
            ylabel = axes[0]
        if not xlabel and axes and len(axes) > 1:
            xlabel = axes[1]
        array2show = self.get_display()
        return show_array(array2show, cut_wings, title, xlabel, ylabel, self.data_unit(), **kwargs)



# __oOo__


import numpy as np

__version__= "$Id: $"




def computeRadialProfile(image, centerInPxY, centerInPxX):
    yCoord, xCoord= np.indices(image.shape)
    yCoord= (yCoord - centerInPxY)
    xCoord= (xCoord - centerInPxX)
    rCoord=np.sqrt(xCoord**2 + yCoord**2)
    indexR= np.argsort(rCoord.flat)
    radialDistancesSorted= rCoord.flat[indexR]
    imageValuesSortedByRadialDistance= image.flat[indexR]
    integerPartOfRadialDistances= radialDistancesSorted.astype(np.int64)
    deltaRadialDistance= integerPartOfRadialDistances[1:] - \
        integerPartOfRadialDistances[:-1]
    radialDistanceChanges= np.where(deltaRadialDistance)[0]
    nPxInBinZero= radialDistanceChanges[0]+ 1
    nPxInRadialBin= radialDistanceChanges[1:] - \
        radialDistanceChanges[:-1]
    imageRadialCumSum= np.cumsum(imageValuesSortedByRadialDistance,
                                 dtype=np.float64)
    imageSumInBinZero= imageRadialCumSum[radialDistanceChanges[0]]
    imageSumInRadialBin= \
        imageRadialCumSum[radialDistanceChanges[1:]] - \
        imageRadialCumSum[radialDistanceChanges[:-1]]
    profileInZero= imageSumInBinZero / nPxInBinZero
    profileFromOne= imageSumInRadialBin / nPxInRadialBin
    profile= np.hstack([profileInZero, profileFromOne])

    distanceRadialCumSum= np.cumsum(radialDistancesSorted)
    distanceSumInBinZero= distanceRadialCumSum[radialDistanceChanges[0]]
    distanceSumInRadialBin= \
        distanceRadialCumSum[radialDistanceChanges[1:]] - \
        distanceRadialCumSum[radialDistanceChanges[:-1]]
    distanceInZero= distanceSumInBinZero / nPxInBinZero
    distanceFromOne= distanceSumInRadialBin / nPxInRadialBin
    radialDistance= np.hstack([distanceInZero, distanceFromOne])
    return profile, radialDistance




'''
Utility functions strictly reserved for dataelab classes
'''
def is_dataelab(x):
    '''Return whether x is supports the dataelab protocol'''
    return hasattr(x, 'get_data') and hasattr(x, 'astropy_unit')



import numpy as np


class Indexer(object):

    def __init__(self):
        self.dicCoord = {'x': 0, 'y': 1}

    def _xy_index(self, coord):
        if not hasattr(coord, '__iter__'):
            coord = [coord]
        indexCoord = [self.dicCoord[x] for x in coord]
        return indexCoord

    def interleaved_xy(self, *args, **kwargs):
        '''
        default: coord
        Accepted keywords:
        coord= x and/or y
        '''
        acceptedKeywords = ['coord', 'coords', 'axis']
        for k, v in kwargs.items():
            if k not in acceptedKeywords:
                raise Exception('possible keywords are ' +
                                str(acceptedKeywords))
            if v == 'x':
                return slice(0, None, 2)
            elif v == 'y':
                return slice(1, None, 2)
            else:
                raise Exception("%s not in accepted"
                                " values ('x', 'y')" % (v))

        # Since we are here, no keywords were specified
        if len(args) == 0:
            return slice(0, None, 1)
        elif len(args) == 1:
            if args[0] == 'x':
                return slice(0, None, 2)
            elif args[0] == 'y':
                return slice(1, None, 2)
            elif isinstance(args[0], int):
                return slice(args[0], args[0]+1)
        else:
            raise Exception("%s not in accepted"
                            " values ('x', 'y')" % str(v))

    def sequential_xy(self, maxindex, *args, **kwargs):
        '''
        default: coord
        Accepted keywords:
        coord= x and/or y
        '''
        acceptedKeywords = ['coord', 'coords', 'axis']
        for k, v in kwargs.items():
            if k not in acceptedKeywords:
                raise Exception('possible keywords are ' +
                                str(acceptedKeywords))
            if v == 'x':
                return slice(0, maxindex//2)
            elif v == 'y':
                return slice(maxindex//2, None)
            else:
                raise Exception("%s not in accepted"
                                " values ('x', 'y')" % (v))

        # Since we are here, no keywords were specified
        if len(args) == 0:
            return slice(0, None, 1)
        elif len(args) == 1:
            if args[0] == 'x':
                return slice(0, maxindex//2)
            elif args[0] == 'y':
                return slice(maxindex//2, None)
            elif isinstance(args[0], int):
                return slice(args[0], args[0]+1)
        else:
            raise Exception("%s not in accepted"
                            " values ('x', 'y')" % str(v))
        
    def xy(self, *args, **kwargs):
        '''
        default: coord
        Accepted keywords:
        coord= x and/or y
        '''
        acceptedKeywords = ['coord', 'coords', 'axis']
        coord = ['x', 'y']
        if len(args) > 0:
            coord = args[1]
        for k, v in kwargs.items():
            if k not in acceptedKeywords:
                raise Exception('possible keywords are ' +
                                str(acceptedKeywords))
            if k == 'coord' or k == 'coords' or k == 'axis':
                coord = v
        return self._xy_index(coord)

    # TODO add boundary checks with maxrange on args
    @staticmethod
    def myrange(*args, maxrange):
        
        if len(args)==0:
            return range(maxrange)
        if len(args)==1:
            return range(args[0])
        if len(args)==2:
            return range(args[0], args[1])
        if len(args)==3:
            return range(args[0], args[1], args[2])
        raise ValueError
            
class ModeIndexer(Indexer):

    def __init__(self, max_mode=None):
        super().__init__()
        self._max_mode = max_mode        

    def modes(self, *args, **kwargs):
        '''
        default: all modes
        mode = single mode
        modes = list of modes
        from_mode = first mode
        to_mode = last mode
        '''
        modes = None
        from_mode = None
        to_mode = None
        if (len(args) > 0) & (len(args) < 2):
            modes = args[0]
        for k, v in kwargs.items():
            if k == 'modes' or k == 'mode':
                modes = v
            if k == 'from_mode':
                from_mode = v
            if k == 'to_mode':
                to_mode = v
        if modes is not None:
            return modes
        if from_mode is None:
            from_mode = 0
        if to_mode is None:
            to_mode = self._max_mode
        if to_mode is None:
            raise Exception("to_mode was not specified and max_mode was not set "
                            "when initializing the ModeIndexer, cannot continue") 
        return np.arange(from_mode, to_mode)


class DefaultIndexer(Indexer):

    def __init__(self):
        super().__init__()

    def elements(self, *args, **kwargs):
        '''
        default: all elements
        element = single element
        elements = list of elements
        from_element = first element
        to_element = last element
        '''
        elements = None
        from_element = None
        to_element = None
        if (len(args) > 0) & (len(args) < 2):
            elements = args[0]
        for k, v in kwargs.items():
            if k == 'elements' or k == 'element':
                elements = v
            if k == 'from_element':
                from_element = v
            if k == 'to_element':
                to_element = v
        if elements is not None:
            return elements
        return slice(from_element, to_element)


class RowColIndexer(Indexer):

    def __init__(self):
        super().__init__()

    def rowcol(self, *args, **kwargs):
        '''
        rows = index
        cols = index
        '''
        rows = slice(None, None, None)
        cols = slice(None, None, None)
        if len(args) == 1:
            rows = args[0]
        elif len(args) == 2:
            rows, cols = args
        elif len(args) > 2:
            raise IndexError('Unsupported rowcol index')
        
        accepted = 'rows cols row_from col_from row_to col_to row_step col_step'.split()
        for kw in kwargs:
            if kw not in accepted:
                raise IndexError(f'Unsupported index keyword {kw}')

            if kw == 'rows':
                rows = kwargs['rows']
            if kw == 'cols':
                cols = kwargs['cols']
            if kw == 'row_from':
                rows = slice(kwargs['row_from'], rows.stop, rows.step)
            if kw == 'col_from':
                cols = slice(kwargs['col_from'], cols.stop, cols.step)
            if kw == 'row_to':
                rows = slice(rows.start, kwargs['row_to'], rows.step)
            if kw == 'col_to':
                cols = slice(cols.start, kwargs['col_to'], cols.step)
            if kw == 'row_step':
                rows = slice(rows.start, rows.stop, kwargs['row_step'])
            if kw == 'col_step':
                cols = slice(cols.start, cols.stop, kwargs['col_step'])

        return (rows, cols)



def logFailureAndReturnNotAvailable(func):

    def wrappedMethod(self, *args, **kwds):
        try:
            return func(self, *args, **kwds)
        except Exception as e:
            self._logger.warn("'%s' failed: %s" % (
                func.__name__, str(e)))
            return NotAvailable()

    return wrappedMethod


class CanBeIncomplete(object):
    _verbose = False

    def __getattr__(self, attr):
        if self._verbose:
            print('%s %s __getattr__ %s' % (self.__class__.__name__,
                                            hex(id(self)), attr))
        iPythonSpecials = ['trait_names',
                           '_getAttributeNames',
                           '__length_hint__']
        if attr in iPythonSpecials:
            if self._verbose:
                print('iPython completer I got you!')
            raise AttributeError
    
        # Numpy array casting: will result in an empty array
        if attr.startswith('__array'):
            raise AttributeError

        return NotAvailable()


class NotAvailable(CanBeIncomplete):
    _verbose = False

    def __init__(self):
        if self._verbose:
            print('%s __init__ %s' % (hex(id(self)),
                                      self.__class__.__name__))

    def __getitem__(self, key):
        if self._verbose:
            print('%s __getitem__ %s' % (hex(id(self)), str(key)))
        return self

    def __setitem__(self, key, value):
        if self._verbose:
            print('__setitem__ %s %s' % (str(key), str(value)))
        pass

    def __call__(self, *args, **kwargs):
        if self._verbose:
            print('%s __call__ %s %s' % (hex(id(self)),
                                         str(args), str(kwargs)))
        return self

    def __iter__(self):
        return self

    def next(self):          # Python2
        raise StopIteration

    def __next__(self):      # Python3
        raise StopIteration

    def __repr__(self):
        return 'NA'

    def __len__(self):
        return 0

    def __add__(self, other):
        return self

    def __sub__(self, other):
        return self

    def __mul__(self, other):
        return self

    def __matmul__(self, other):
        return self

    def __floordiv__(self, other):
        return self

    def __mod__(self, other):
        return self

    def __divmod__(self, other):
        return self

    def __pow__(self, other):
        return self

    def __lshift__(self, other):
        return self

    def __rshift__(self, other):
        return self

    def __and__(self, other):
        return self

    def __xor__(self, other):
        return self

    def __or__(self, other):
        return self

    def __div__(self, other):
        return self

    def __truediv__(self, other):
        return self

    def __radd__(self, other):
        return self

    def __rsub__(self, other):
        return self

    def __rmul__(self, other):
        return self

    def __rmatmul__(self, other):
        return self

    def __rdiv__(self, other):
        return self

    def __rtruediv__(self, other):
        return self

    def __rfloordiv__(self, other):
        return self

    def __rmod__(self, other):
        return self

    def __rdivmod__(self, other):
        return self

    def __rpow__(self, other):
        return self

    def __rlshift__(self, other):
        return self

    def __rrshift__(self, other):
        return self

    def __rand__(self, other):
        return self

    def __rxor__(self, other):
        return self

    def __ror__(self, other):
        return self

    def __iadd__(self, other):
        return self

    def __isub__(self, other):
        return self

    def __imul__(self, other):
        return self

    def __imatmul__(self, other):
        return self

    def __idiv__(self, other):
        return self

    def __itruediv__(self, other):
        return self

    def __ifloordiv__(self, other):
        return self

    def __imod__(self, other):
        return self

    def __ipow__(self, other):
        return self

    def __ilshift__(self, other):
        return self

    def __irshift__(self, other):
        return self

    def __iand__(self, other):
        return self

    def __ixor__(self, other):
        return self

    def __ior__(self, other):
        return self

    def __neg__(self):
        return self

    def __pos__(self):
        return self

    def __abs__(self):
        return self

    def __invert__(self):
        return self

    def __format__(self, fmt):
        return 'NA'

    @staticmethod
    def transformInNotAvailable(obj):
        obj.__class__ = NotAvailable

    @staticmethod
    def isNotAvailable(obj):
        return isinstance(obj, NotAvailable)



# -*- coding: utf-8 -*-
#########################################################
#
# who       when        what
# --------  ----------  ---------------------------------
# apuglisi  2019-09-15  Created
#
#########################################################

import numpy as np

def circular_mask(size, center=None, diameter=1.0, obstruction=0.0):
    '''
    Creates a circular mask for *size*x*size* elements, where elements
    whose distance from *center* is higher than *ob/2* and smaller
    than *diameter/2* are set to one, and all others at zero.
    Both *diameter* and *obstruction* are in percentage (0..1),
    where 1 is the full aperture.
    *center* should be a sequence of two coordinates. If omitted or None,
    it will be set to [*size/2*,*size/2*].
    
    See also arte.types.mask for the class CircularMask and derived classes
    '''
    if center is None:
        center = [(size-1)/2, (size-1)/2]

    xx, yy = np.ogrid[:size,:size]
    dist = np.sqrt((xx - center[0])**2 + (yy-center[1])**2)

    mask = dist <= diameter*size/2
    mask *= dist >= obstruction*size/2
    return mask

# ___oOo___



import numpy as np
from arte.types.mask import CircularMask

TPS_RBF='tps_rbf'
GAUSS_RBF='gauss_rbf'
INV_QUADRATIC='inv_quadratic'
MULTIQUADRIC='multiquadric'
INV_MULTIQUADRIC='inv_multiquadric'



class RBFGenerator(object):

    #documentation for this class
    '''
    This class generates a set of radial basis functions (RBF) from a given set of coordinates.
    The radial basis functions are generated using a given RBF function. The RBF functions are
    defined as follows:
    - TPS_RBF: r * log(r)
    - GAUSS_RBF: exp(-(eps*r)**2)
    - INV_QUADRATIC: 1/(1+(eps*r)**2)
    - MULTIQUADRIC: sqrt(1+(eps*r)**2)
    - INV_MULTIQUADRIC: 1/sqrt(1+(eps*r)**2)

    The RBF functions are computed over a boolean mask that defines the pupil. The mask can be
    a CircularMask object or a scalar value representing the radius of the pupil. In the latter
    case, the mask is generated internally.

    Parameters
    ----------
    pupil: real or `~arte.types.mask.CircularMask`
        If a scalar value, the argument is used as pupil diameter in pixels.
        If a `~arte.types.mask.CircularMask`, the argument is used as mask
        representing the unit disk.
    coords: list of tuples
        List of tuples containing the coordinates of the RBF functions.
    rbfFunction: str
        String representing the RBF function to use. The available functions are:
        - TPS_RBF
        - GAUSS_RBF
        - INV_QUADRATIC
        - MULTIQUADRIC
        - INV_MULTIQUADRIC
    eps: real
        Value of the epsilon parameter for the RBF function.

    Examples
    --------
    Create a set of radial basis functions using a set of coordinates and the TPS_RBF function
    >>> from arte.utils.rbf_generator import RBFGenerator
    >>> from arte.types.mask import CircularMask
    >>> import numpy as np
    >>> import matplotlib.pyplot as plt
    >>> coords = [(10,10), (20,20), (30,30)]
    >>> mask = CircularMask((128,128), 64, (64,64))
    >>> rbf = RBFGenerator(mask, coords, 'TPS_RBF')
    >>> rbf.generate()
    >>> rbfCube = rbf.getRBFCube()
    >>> plt.imshow(rbfCube[:,:,0])
    '''

    def __init__(self, pupil, coords, rbfFunction="TPS_RBF", eps=1.0):
        
        if isinstance(pupil, CircularMask):
            self._radius = pupil.radius()
            self._shape = pupil.shape()
            self._center = pupil.center()
            self._boolean_mask = pupil.mask()
        else:
            self._radius = pupil / 2
            sz = np.ceil(pupil)
            self._shape = (sz, sz)
            self._center = np.ones(2) * (sz / 2)
            cm = CircularMask(
                self._shape, maskCenter=self._center, maskRadius=self._radius)
            self._boolean_mask = cm.mask()

        self._fcnDict = {'TPS_RBF': 'tps_rbf',
                         'GAUSS_RBF': 'gauss_rbf',
                         'INV_QUADRATIC': 'inv_quadratic',
                         'MULTIQUADRIC': 'multiquadric',
                         'INV_MULTIQUADRIC': 'inv_multiquadric'}
        self._dictCoords = {i: coord for i, coord in enumerate(coords)}
        self._rbfFunction = self._fcnDict[rbfFunction]
        self._eps = eps
        self._nmodes = 0
        self._rbfBase = None
        self._dictCache = {}

    def generate(self):
        self._nmodes = len(self._dictCoords) 
        nPxY = self._boolean_mask.shape[0]
        nPxX = self._boolean_mask.shape[1]

        for i in range(self._nmodes):
            y0,x0 = self._dictCoords[i]
            cc = np.expand_dims((x0,y0), axis=(1, 2))
            Y, X = (np.mgrid[0.5: nPxY + 0.5: 1,
                            0.5: nPxX + 0.5: 1] - cc) / self._radius
            r = np.sqrt(X ** 2 + Y ** 2)
            if i not in self._dictCache:
                #print("add mode %d" % i)
                if self._rbfFunction == 'tps_rbf':
                    cdata = self.tps_rbf(r)
                elif self._rbfFunction == 'gauss_rbf':
                    cdata = self.gauss_rbf(r, self._eps)
                elif self._rbfFunction == 'inv_quadratic':
                    cdata = self.inv_quadratic(r, self._eps)
                elif self._rbfFunction == 'multiquadric':
                    cdata = self.multiquadric(r, self._eps)
                elif self._rbfFunction == 'inv_multiquadric':
                    cdata = self.inv_multiquadric(r, self._eps)
                else:
                    raise ValueError("Invalid RBF function %s" % self._rbfFunction)
            self._dictCache[i] = np.ma.masked_array(data=cdata, mask=self._boolean_mask)
        self._rbfBase = np.rollaxis(np.ma.masked_array([self.getRBF(n) 
                                    for n in range(self._nmodes)]),0,3)

    def getRBF(self, index):
        return self._dictCache[index]

    def getRBFDict(self, index):
        return {i: self._rbfBase[:, :, i] for i in index}

    def getRBFCube(self):
        return self._rbfBase
    
    def tps_rbf(self, r):
        return r * np.log(r**r)

    def gauss_rbf(self, r, eps):
        return np.exp(-(eps*r)**2)

    def inv_quadratic(self, r,eps):
        return 1/(1+(eps*r)**2)

    def multiquadric(self, r,eps):
        return np.sqrt(1+(eps*r)**2)

    def inv_multiquadric(self, r,eps):
        return 1/np.sqrt(1+(eps*r)**2)

    def getModesDict(self, indexVector):
        return self.getRBFDict(indexVector)

    def first_mode(self):
        return 0




"""
A well-behaved writer class.

It is important to help readability. Therefore, arte has code conventions.

**Rules:**
 - Line indentation is 4 spaces. No Tabs!
 - Maximum line length is 80 characters.
 - Comments should be complete sentences in English.
 - Modules use short, all-lowercase names with underscores (snake_case).
 - Class names use the UpperCamelCase convention.
 - Method names use all-lowercase with underscore (snake_case) convention.
 - Docstrings use the numpydoc convention: https://numpydoc.readthedocs.io/


**Eclipse Support**

 - The Python plugin PyDev for Eclipse has a tool for checking source code
   on conformity with PEP8 standard. Use it!
 - Set Pydev->Editor->Code Style->Code Formatter->Formatter style=Pydev.Formatter
 - Check Pydev->Editor->Code Analysis->Do Code Analysis

"""

import os


class NiceWriter:
    """
    A nice writer.

    This class is an example of code styling conventions.

    You can do some math:

    .. math::

        f(x) = x^{-5/3} + \\frac{1}{L_0^2}^{-42}


    Arguments
    ---------
    path: str
        the `path` of something


    Examples
    -------

    Show that 2+3=5

    >>> a = 2
    >>> b = 3
    >>> print(a+b)
    5

    """

    A_CLASS_CONSTANT = 'whatever it is, centralize constant here'
    NUMBER_OF_SLOPES = 1600  # Not N_SLOPES. Be descriptive

    def __init__(self, path):
        self._path = path

    def public_work(self, an_argument, an_array, a_keyword='not needed'):
        """
        One-liner description here.

        More detailed description here, using reStructuredText if needed.

        Parameters
        ----------
        an_argument : str
           argument description here.

        an_array: numpy.array
            pass `an_array` as an array

        a_keyword: str, default='not needed'
           something optional.

        Returns
        -------
        k : int
           whatever.

        """
        if self._path.startswith(an_argument):
            os.chdir(self.A_CLASS_CONSTANT)
        else:
            os.remove(self._path)

        for i in range(0, 3):
            self._say_something_private(i)

        j = 5
        k = 3
        a_container = []
        while k < 9:
            a_container.append(j + k)
            k = k + 1

        return sum(a_container)

    def _say_something_private(self, x='love you'):
        """
        A private method.

        A leading underscore denotes private methods.

        Actually you don't need to docstring private methods.
        This one is not rendered in documentation


        """
        print(x)


class InafIsAnAcronym():
    '''
    InafIsAnAcronym is more readable than INAFIsAnAcronym

    So use Inaf, Lbt, Eso and so on in acronyms
    '''
    pass



'''
Created on 13 mar 2020

@author: giuliacarla
'''

import numpy as np


class LaplaceTransferFunction():
    """
    Return Open Loop Transfer Function(OLTF), Rejection Transfer Function (RTF)
    and Noise Transfer Function (NTF) of a basic control system made of a sensor with
    integration time T and readout time Tr, an integral control with gain G and 
    delay Tc, and sample-and-hold commands to an ideal actuator.
    The TFs are computed using Laplace transform.

    Parameters
    ----------
    temporal_freqs: `~numpy.ndarray`
        Array of temporal frequencies used to compute OLTF, RTF and NTF.

    gain: int or float
        System's gain set by the user. If needed, the system's optimal gain
        can be computed (see module 'set_optimal_gain').

    t_integration: int or float
        Time interval over which the system's sensor integrates the input
        signal.
        It determines the system's loop frequency (f_loop = 1 / t_integration).

    t_readout: int or float
        Sensor's readout time. Default is 0.

    t_control: int or float
        Delay introduced by the digital computation of the signal,
        the filtering and the final analogic conversion of the signal.
        Default is 0.
    """

# TODO: restituire dei tipi 'Transfer Function'?

    def __init__(self, temporal_freqs, t_integration,
                 t_readout=0, t_control=0, gain=1):
        self.temporal_freqs = temporal_freqs
        self.T = t_integration
        self.Tr = t_readout
        self.Tc = t_control
        self.gain = gain
        self.Tdelay = t_readout + t_control
        self._s = 1j * 2 * np.pi * temporal_freqs
        self._chooseIntegrator = False
        self._chooseSampleAndHold = False

    def set_integrator(self, true_false):
        self._chooseIntegrator = true_false

    def set_sample_and_hold(self, true_false):
        self._chooseSampleAndHold = true_false

    def set_optimal_gain(self):
        """
        Find the system's optimal gain based on OLTF's amplitude value
        at the frequency where the phase is equal to 135 degrees.
        """
        amp = self.get_amplitude(self.OLTF())
        amp135 = amp[self.idx_omega_135]
        rounding = 0.001
        if amp135 <= 1 + rounding:
            self.gain = self.gain / amp135
        else:
            raise Exception('Input gain is too high. The system is unstable.')

    def get_bandwidth_at_optimal_gain(self):
        return self.temporal_freqs[self.idx_omega_135]

    def rejection_bandwidth(self):
        self._findRejectionBandOmegaIdx()
        return self.temporal_freqs[self.idx_omega_rejection]

    def _findRejectionBandOmegaIdx(self):
        amp = self.get_amplitude(self.RTF())
        idxMax = np.argwhere(amp == amp.max())[0][0]
        ampToMax = amp[:idxMax]
        diff = abs(ampToMax - 1)
        self.idx_omega_rejection = np.argwhere(diff == diff.min())[0][0]

    def _findOmega135Idx(self):
        phi = self.get_phase(self.OLTF(), unit='deg')
        diff = abs(phi - (-135))
        return np.argwhere(diff == diff.min())[0][0]

    @property
    def idx_omega_135(self):
        return self._findOmega135Idx()

    def OLTF(self):
        """
        Get Open Loop Transfer Function.

        Return
        ------
        oltf: `~numpy.ndarray`
        """

        S = self._computeSensorTransferFunction()
        F = self._computeFilterTransferFunction()
        H = self._computeSampleAndHoldTransferFunction()
        A = self._computeActuatorTransferFunction()
        oltf = S * F * H * A
        self._checkFrequencySampling(oltf)
        return oltf

    def RTF(self):
        """
        Get Rejection Transfer Function.

        Return
        ------
        rtf: `~numpy.ndarray`
        """
        return 1. / (1 + self.OLTF())

    def NTF(self):
        """
        Get Noise Transfer Function

        Return
        ------
        ntf: `~numpy.ndarray`
        """
        return self.OLTF() / (1 + self.OLTF())

    def get_amplitude(self, transfer_function):
        """
        Get transfer function's amplitude.

        Parameters
        ----------
        transfer_function: `~numpy.ndarray`

        Return
        ------
        amplitude: `~numpy.ndarray`
            Transfer function's module.
        """

        amplitude = abs(transfer_function)
        return amplitude

    def get_phase(self, transfer_function, unit='deg'):
        """
        Get transfer function's phase.

        Parameters
        ----------
        transfer_function: `~numpy.ndarray`
        unit: str
            Phase's unit. Default is 'radians'.

        Return
        ------
        phi: `~numpy.ndarray`
            Transfer function's phase.

        unit: str
            Phase's unit. It can be 'radians' or 'deg'. Default is 'deg'.
        """
        phi = np.unwrap(2 * np.angle(transfer_function)) / 2
        if unit == 'radians':
            return phi
        elif unit == 'deg':
            phi = np.rad2deg(phi)
            return phi

    def _checkFrequencySampling(self, transfer_function):
        import warnings

        phiInRad = self.get_phase(transfer_function)
        dPhi = np.diff(phiInRad)
        jump_up = np.where(dPhi > 0)
        if jump_up[0].shape[0] != 0:
            if np.diff(jump_up).min() == 1:
                warnings.warn(
                    'You may have to increase the frequency sample rate.')

    def _computeSensorTransferFunction(self):
        return (1 - np.exp(-self._s * self.T)) / (
            self._s * self.T) * np.exp(-self._s * self.Tr)

    def _computeFilterTransferFunction(self):
        if self._chooseIntegrator is False:
            F = self.gain
        elif self._chooseIntegrator is True:
            F = self.gain / self._s
        return F

    def _computeSampleAndHoldTransferFunction(self):
        if self._chooseSampleAndHold is False:
            H = 1
        elif self._chooseSampleAndHold is True:
            H = (1 - np.exp(-self._s * self.T)) / self._s * np.exp(
                -self._s * self.Tc)
        return H

    def _computeActuatorTransferFunction(self):
        # Per ora solo attuatore perfetto.
        return 1


class ZetaTransferFunction():
    """
    Get OLTF, RTF and NTF of a control system using Zeta transform.

    Parameters
    ----------
    f_loop:

    n_iterations:

    gain:

    delay:

    """

    def __init__(self, f_loop, n_iterations, gain, delay):
        self.loop_frequency = f_loop
        self.n_iter = n_iterations
        self.gain = gain
        self.delay = delay
        self.set_temporal_frequency_array()

    def set_temporal_frequency_array(self, n_points=None):
        f_min = self.loop_frequency / self.n_iter
        f_nyquist = self.loop_frequency / 2
        if n_points is None:
            n_points = int(self.n_iter / 2)
        self.temporal_freqs = np.linspace(f_min, f_nyquist, n_points)
        self.z = np.exp(
            1j * 2 * np.pi * self.temporal_freqs / self.loop_frequency)

    def get_amplitude(self, tf):
        return abs(tf)

    def get_phase(self, tf, unit='radians'):
        phi = np.unwrap(2 * np.angle(tf)) / 2
        if unit == 'radians':
            return phi
        elif unit == 'deg':
            return np.rad2deg(phi)

    def RTF(self):
        rtf = (1 - self.z ** (-1)) / (
            1 - self.z ** (-1) + self.gain * self.z ** (-self.delay))
        return rtf

    def NTF(self):
        ntf = -(self.gain * self.z ** (-self.delay)) / (
            1 - self.z ** (-1) + self.gain * self.z ** (-self.delay))
        return ntf


class IdealTransferFunction():

    def __init__(self, rtf=0, ntf=-1):
        self.rtf = rtf
        self.ntf = ntf

    def RTF(self):
        return self.rtf

    def NTF(self):
        return self.ntf



import numpy as np


class Wavefront(object):

    def __init__(self, wf, counter=0):
        self._wf = wf
        self._counter = counter

    def toNumpyArray(self):
        return self._wf

    @staticmethod
    def fromNumpyArray(wfAsNumpyArray, counter=0):
        # assert isinstance(wfAsNumpyArray, np.ma.masked_array)
        return Wavefront(wfAsNumpyArray, counter)

    def counter(self):
        return self._counter

    def setCounter(self, counter):
        self._counter = counter

    def __eq__(self, o):
        if self._counter != o._counter:
            return False
        if not np.array_equal(self._wf, o._wf):
            return False
        return True

    def __ne__(self, o):
        return not self.__eq__(o)

    def std(self):
        return self._wf.std()




# -*- coding: utf-8 -*-
#########################################################
#
# who       when        what
# --------  ----------  ---------------------------------
# apuglisi  2019-09-08  Created
#
#########################################################
from __future__ import print_function

import os.path
import tempfile
import subprocess
import numpy as np
from scipy.io import readsav
from collections.abc import Iterable


def compareIDL(idlscript, pyscript, vars_to_compare,
               precision=1e-5, verbose=0, tmpfile=None):
    '''
    Compare IDL and Python routines results, to aid in porting.

    This function will run an IDL batch script (containing statements
    that will be executed as if they were typed on the IDL command prompt)
    and a Python script. After both scripts have been run, the variables
    listed in vars_to_compare are extracted from both sessions and
    compared numerically to the specified precision limit.

    Parameters
    ----------
    idlscript: string or list of strings
        IDL statements. If in a single string,
        they must be separated by newlines
    pyscript: string or list of strings
        Python statements. If in a single string,
        they must be separated by newlines
    vars_to_compare: list of strings
        variable names to compare
    precision: float, optional
        relative numerical precision
    verbose: int, optional
        verbose level: if greater than zero, will print on stdout the list
        of variables for which the comparison fails; If greater than one,
        will also print the variable values.
    tmpfile: string, optional
        filename for the temporary file used to save IDL variables.
        If None, a default filename in the system's temporary
        directory will be used.

    Returns
    -------
    bool
        True if the comparison is within the precision limit, False otherwise
    '''

    if tmpfile is None:
        tmpfile = os.path.join(tempfile.gettempdir(), 'idl_compare.sav')

    savecmd = ','.join(['\nSAVE',
                        *vars_to_compare,
                        'FILENAME="%s"\n' % tmpfile])

    if isinstance(idlscript, Iterable) and not isinstance(idlscript, str):
        idlscript = '\n'.join(idlscript)

    if isinstance(pyscript, Iterable) and not isinstance(pyscript, str):
        pyscript = '\n'.join(pyscript)

    p = subprocess.Popen('idl', stdin=subprocess.PIPE, shell=True)
    p.communicate((idlscript + savecmd).encode())
    idldata = readsav(tmpfile)

    exec(pyscript)

    good = []
    for varname in vars_to_compare:
        idlvar = idldata[varname]
        pyvar = locals()[varname]

        if pyvar is None:
            print('%s is None, cannot compare' % varname)
            return False

        goodvar = np.all(np.isclose(idlvar, pyvar))
        if verbose and not goodvar:
            print('%s differs' % varname)
            if verbose > 1:
                print('IDL value   : ', idlvar)
                print('Python value: ', pyvar)
        good.append(goodvar)

    return all(good)

# ___oOo___



import numpy as np
from arte.types.region_of_interest import RegionOfInterest
from arte.utils.image_moments import ImageMoments
from skimage import feature
from skimage import measure, draw
from scipy import optimize
import warnings
import matplotlib.pyplot as plt


class BaseMask():
    '''
    '''

    def __init__(self, mask_array):
        self._shape = mask_array.shape
        self._mask = mask_array
    
    def mask(self):
        '''
        Boolean mask of the mask

        Returns
        -------
        mask: boolean `~numpy.array`
            mask of the pupil. Array is True outside the pupil,
            and False inside the pupil
        '''
        return self._mask
    
    def as_masked_array(self):
        return np.ma.array(np.ones(self._shape),
                           mask=self.mask())
        
    def shape(self):
        '''
        Array shape

        Returns
        -------
        shape: list (2,)
            shape of the mask array
        '''
        return self._shape
    
    @staticmethod
    def from_masked_array(numpy_masked_array):
        return BaseMask(numpy_masked_array)
    
    def __eq__(self, other):
        if isinstance(other, BaseMask):
            return np.array_equal(self.mask(), other.mask())
        return False

    def __hash__(self):
        return hash(self.mask().tobytes())

    # TODO: funzione per verificare che tutti i punti di questa maschera stanno
    # dentro una maschera passata


class CircularMask(BaseMask):
    '''
    Represent a circular mask

    A `~numpy.array` representing a circular pupil. Frame shape, pupil radius
    and center can be specified.

    Use `mask` method to access the mask as boolean mask (e.g. to be used
    in a `~numpy.ma.masked_array` object) with False values where the frame is 
    not masked (i.e. within the pupil) and True values outside

    Use `asTransmissionValue` method to acces the mask as transmission mask, i.e. 
    1 within the pupil and 0 outside. Fractional transmission for edge pixels
    is not implemented

    If a `~numpy.ma.masked_array` having a circular mask is available, the static
    method `fromMaskedArray` can be used to create a `CircularMask` object having
    the same shape of the masked array and the same pupil center and radius


    Parameters
    ----------
        frameShape: tuple (2,)
            shape of the returned array

        maskRadius: real
            pupil radius in pixel

        maskCenter: list (2,) or `~numpy.array`
            Y-X coordinates of the pupil center in pixel
    '''

    FITTING_METHOD_IMAGE_MOMENTS = "ImageMoments"
    FITTING_METHOD_RANSAC = "RANSAC"
    FITTING_METHOD_CENTER_OF_GRAVITY = "COG"
    FITTING_METHOD_CORRELATION = "correlation"

    def __init__(self,
                 frameShape,
                 maskRadius=None,
                 maskCenter=None):
        self._shape = frameShape
        self._maskRadius = maskRadius
        self._maskCenter = maskCenter
        self._mask = None
        self._computeMask()

    def __repr__(self):
        return "shape %s, radius %f, center %s" % (
            self._shape, self._maskRadius, self._maskCenter)

    def _computeMask(self):
        if self._maskRadius is None:
            self._maskRadius = min(self._shape) / 2.
        if self._maskCenter is None:
            self._maskCenter = 0.5 * np.array([self._shape[0],
                                               self._shape[1]])

        r = self._maskRadius
        cc = self._maskCenter
        y, x = np.mgrid[0.5: self._shape[0] + 0.5:1,
                        0.5: self._shape[1] + 0.5:1]
        self._mask = np.where(
            ((x - cc[1]) ** 2 + (y - cc[0]) ** 2) <= r ** 2, False, True)

    def asTransmissionValue(self):
        '''
        Mask as a transmission mask: 1 for non-masked elements,
        0 for masked elements.

        Returns
        -------
        transmission_value: ndarray[int]
            transmission mask as a numpy array with dtype int
        '''
        return np.logical_not(self._mask).astype(int)
    
    #def as_masked_array(self):
    #    return np.ma.array(np.array(self.asTransmissionValue(), dtype=float),
    #                       mask=self.mask())

    def radius(self):
        '''
        Radius of the mask 

        Returns
        -------
        radius: real
            mask radius in pixel

        '''
        return self._maskRadius

    def center(self):
        '''
        Y, X coordinates of the mask center

        Returns
        -------
        center: `~numpy.array` of shape (2,)
            Y, X coordinate of the mask center in the array reference system

        '''
        return self._maskCenter

    @staticmethod
    def fromMaskedArray(maskedArray, method='ImageMoments', **keywords):
        '''
        Creates a `CircularMask` roughly corresponding to the mask of a masked
        array
        
        Returns a `CircularMask` object having radius and center guessed from
        the passed mask using `ImageMoments` centroid and semiAxes as default.
        
        Important note: the created `CircularMask` object using "ImageMoments" 
        is not guaranteed to have the same set of valid points of the 
        passed mask, but it is included in the passed mask, i.e. all the 
        valid points of the created mask are also valid points of the passed 
        masked array. For the other methods this last property is not guaranteed.  

        Other valid circle estimation methods are:
        - "RANSAC" : circle fitting using RANSAC algorithm on Canny edge detection
        - "COG" : circle fitting using Center of gravity of the mask
        - "correlation" : circle fitting using correlation of the mask with a disk template

        
        Parameters
        ----------
        maskedArray: `~numpy.ma.MaskedArray` 
            a masked array with a circular mask
        method: string
            method for circle estimation. Default is "ImageMoments"

        Returns
        -------
        circular_mask: `CircularMask`
            a circular mask included in the mask of `maskedArray`

        '''
        assert isinstance(maskedArray, np.ma.masked_array)
        shape = maskedArray.shape
        if method == CircularMask.FITTING_METHOD_IMAGE_MOMENTS:
            again = 0.995
            while again:
                im = ImageMoments(maskedArray.mask.astype(int) * -1 + 1)
                centerYX = np.roll(im.centroid(), 1)
                radius = again * np.min(im.semiAxes())
                circularMask = CircularMask(shape, radius, centerYX)
                if np.in1d(circularMask.in_mask_indices(),
                           np.argwhere(maskedArray.mask.flatten() == False)).all():
                    again = False
                if radius < 1:
                    raise Exception("Couldn't estimate a CircularMask")
                else:
                    again *= 0.9
        elif method == CircularMask.FITTING_METHOD_RANSAC:
            img = np.asarray(maskedArray.mask.astype(float) * -1 + 1)
            img[img > 0] = 128
            edge = img.copy()
            edge = feature.canny(img, keywords.pop('sigmaCanny', 2))

            coords = np.column_stack(np.nonzero(edge))

            model, inliers = measure.ransac(
                coords, measure.CircleModel,
                min_samples = keywords.pop('min_samples', 20), 
                residual_threshold = keywords.pop('residual_threshold',0.001),
                max_trials = keywords.pop('max_trials',10000))
            cx, cy, r = model.params

            if  keywords.pop('display', False) is True:
                print(r"Cx-Cy {:.2f}-{:.2f}, R {:.2f}".format(cx, cy, r))
                rr, cc = draw.disk((model.params[0], model.params[1]),
                                model.params[2],
                                shape=img.shape)
                img[rr, cc] += 512
                CircularMask._dispnobl(img)

            cy+=0.5
            cx+=0.5
            circularMask = CircularMask(img.shape, r, [cx,cy])

        elif method == CircularMask.FITTING_METHOD_CENTER_OF_GRAVITY:
            img = np.asarray(maskedArray.mask.astype(int) * -1 + 1)
            regions = measure.regionprops(img)
            bubble = regions[0]
            y0, x0 = bubble.centroid
            y0+=0.5
            x0+=0.5
            r = bubble.major_axis_length / 2.
            circularMask = CircularMask(img.shape, r, [y0,x0])
            
        elif method == CircularMask.FITTING_METHOD_CORRELATION:
            
            img = np.asarray(maskedArray.mask.astype(int) * -1 + 1)
            regions = measure.regionprops(img)
            bubble = regions[0]

            x0, y0 = bubble.centroid
            r = bubble.major_axis_length / 2.

            initial_guess = (x0, y0, r)
            display = keywords.pop('display', False)

            if display:
                fign=plt.figure()
                
            def _cost_disk(params):
                x0, y0, r = params
                coords = draw.disk((x0, y0), r, shape=img.shape)
                template = np.zeros_like(img)
                template[coords] = 1
                if display:
                    CircularMask._dispnobl(template + img, fign)
                return -np.sum((template > 0) & (img > 0))

            res = optimize.minimize(_cost_disk, initial_guess,
                                    method='Nelder-Mead', **keywords)
            x0, y0, r = res.x
            y0+=0.5
            x0+=0.5
            if res.success is True:
                circularMask = CircularMask(img.shape, r, [x0,y0])

            else:
                raise Exception("Fit circle didn't converge %s" % res)
        
        else:
            raise ValueError("Unknown method %s" % method)

        if not np.in1d(circularMask.in_mask_indices(),
                       np.argwhere(maskedArray.mask.flatten() == False)).all():
            warnings.warn(
                "The generated CircularMask is not completely included in the passed masked array")
        return circularMask

    def regionOfInterest(self):
        centerX = int(self.center()[1])
        centerY = int(self.center()[0])
        radius = int(self.radius())
        return RegionOfInterest(centerX - radius, centerX + radius,
                                centerY - radius, centerY + radius)

    def in_mask_indices(self):
        return self.asTransmissionValue().flatten().nonzero()[0]

    def _dispnobl(img, fign=None, **kwargs):
        if fign is not None:
            plt.figure(fign.number)
        plt.clf()
        plt.imshow(img, aspect='auto', **kwargs)
        plt.colorbar()
        plt.draw()
        plt.ion()
        plt.show()
        plt.pause(0.001)
        plt.ioff()


class AnnularMask(CircularMask):
    '''
    Inheritance of CircularMask class to provide an annular mask

    Added inRadius parameter, radius of central obstruction.
    Default inRadius values is 0 with AnnularMask converging to CircularMask
    '''

    def __init__(self,
                 frameShape,
                 maskRadius=None,
                 maskCenter=None,
                 inRadius=0):
        self._inRadius = inRadius
        super().__init__(frameShape, maskRadius, maskCenter)

    def __repr__(self):
        return "shape %s, radius %f, center %s, inradius %f" % (
            self._shape, self._maskRadius, self._maskCenter, self._inRadius)

    def inRadius(self):
        return self._inRadius

    def _computeMask(self):

        if self._maskRadius is None:
            self._maskRadius = min(self._shape) / 2.
        if self._maskCenter is None:
            self._maskCenter = 0.5 * np.array([self._shape[0],
                                               self._shape[1]])

        r = self._maskRadius
        cc = self._maskCenter
        y, x = np.mgrid[0.5: self._shape[0] + 0.5:1,
                        0.5: self._shape[1] + 0.5:1]

        tmp = ((x - cc[1]) ** 2 + (y - cc[0]) ** 2) <= r ** 2
        if self._inRadius == 0:
            self._mask = np.where(tmp, False, True)
        else:
            cc = CircularMask(self._shape, self._inRadius, self._maskCenter)
            tmp[cc.asTransmissionValue() > 0] = False
            self._mask = np.where(tmp, False, True)

    @staticmethod
    def fromMaskedArray(maskedArray):
        raise Exception("Not implemented yet. ")



'''Persistent disk cache for immutable method return values.

Disk cache in the style of LBT elab_lib
'''

import os
import pickle
import logging
import tempfile
from functools import wraps, cached_property
from collections import defaultdict
from types import ModuleType

import numpy as np


class TagNotSetException(Exception):
    '''Cache tag has not been set'''


def cache_on_disk(f):
    '''Persistent disk cache for immutable method return values.

    Only use this decorator for methods whose return value never changes.

    Methods decorated with cache_on_disk() will write their return value into
    a npy or pickle file depending on the value type. Subsequent method
    calls will reuse the previous return value, stored in a memory buffer.
    New instances using the same tag will read data from disk on the first call.

    Caching only starts after a global tag has been defined. Call
    set_tag(obj, tag) with obj set the highest object in the hierarchy
    (typically the analyzer instance) to initialize disk caching for all
    inherited and composed (via attributes) objects.

    Tags should be uniquely identifying the dataset to be stored.
    One suggestion is to use both a system identifier (like KAPA or LBTISX)
    and a timestamp, for example LBTISX20240410_112233.

    Call clear_cache(obj) to delete all temporary files for all inherited
    and composed (via attributes) objects. The methods code will be run
    and stored again when called.

    Data is stored in tmpdir/prefix<tag>/filename.npy, where:
     - tmpdir is by default the system temporary directory
     - prefix is by default "cache"
     - tag must be set by the owner class by calling set_tag() at some point
     - "filename" identifies the method and class name

    File always has extension ".npy". Any type different from a numpy array will
    be pickled instead, using the same extension and leveraging numpy's
    transparent object pickling.

    Defaults for tmpdir and prefix can be overriden using set_tmpdir() and
    set_prefix(), each with two arguments:
     - the highest object in the hiearchy (typically the anayzer instance)
     - the new value for tmpdir or prefix
    '''

    # Bind this function to a DiskCacher constructor
    def new_cacher():
        return DiskCacher(f)

    # Access a DiskCacher based on the instance
    @wraps(f)
    def wrapper(self, *args, **kwargs):
        return wrapper.cacher_list[self].execute(self, *args, **kwargs)

    # Build a new DiskCacher whenever a new instance accesses this method
    wrapper.cacher_list = defaultdict(new_cacher)
    return wrapper


def get_disk_cacher(instance, method):
    '''Return the DiskCacher used by *method* for instance *instance*'''
    return method.cacher_list[instance]


def set_tag(root_obj, tag):
    '''Setup disk cache tag for root_obj and child/member objects'''
    for instance, instance_name, cacher_list in _discover_cachers(root_obj):
        cacher_list[instance].set_tag(tag, instance_name)


def clear_cache(root_obj):
    '''Clear cache for root_obj and child/member objects'''
    for instance, _, cacher_list in _discover_cachers(root_obj):
        cacher_list[instance].clear_cache()


def set_tmpdir(root_obj, tmpdir):
    '''Set the directory where cached data is stored'''
    for instance, _, cacher_list in _discover_cachers(root_obj):
        cacher_list[instance].set_tmpdir(tmpdir)


def set_prefix(root_obj, prefix):
    '''Set the prefix for each tag directory'''
    for instance, _, cacher_list in _discover_cachers(root_obj):
        cacher_list[instance].set_prefix(prefix)


_logger = None

def set_logfile(filename, level=logging.DEBUG, name='cache_on_disk'):
    '''Set the file where log output will be written'''
    global _logger
    _logger = logging.getLogger(name)
    _logger.setLevel(level)
    formatter = logging.Formatter('%(asctime)s.%(msecs)03d %(message)s',
                                datefmt='%Y-%m-%d:%H:%M:%S')
    file_handler = logging.FileHandler(filename)
    file_handler.setLevel(level)
    file_handler.setFormatter(formatter)
    _logger.addHandler(file_handler)


def _discover_cachers(obj, objname='root', seen=None):
    '''Generator that yields all DiskCacher objects in all child members'''

    # This set is used to avoid infinite recursion
    # when objects have circular references
    if seen is None:
        seen = set()

    # Exclude properties and cached properties
    properties = ([k for k in dir(obj.__class__)
                   if isinstance(getattr(obj.__class__, k), (property, cached_property))])

    # Some packages like astropy.units define attributes
    # that just raise exceptions if they are accessed (!)
    valid_attrs = {}
    for name in dir(obj):
        if not name in properties:
            try:
                valid_attrs[name] = getattr(obj, name)
            except AttributeError:
                pass

    methods = {k: v for k, v in valid_attrs.items() if callable(v)}
    members = {k: v for k, v in valid_attrs.items() if not callable(v)}

    # Examine both methods and members, because DiskCacher replaces a method
    for name, method in methods.items():
        seen.update([type(method)])
        if hasattr(method, 'cacher_list'):
            yield obj, objname, method.cacher_list

    for name, member in members.items():
        if hasattr(member, 'cacher_list'):
            yield obj, objname, member.cacher_list

        # Recurse into all child members except for special variables.
        if not name.startswith('__') and not type(member) in seen:
            seen.update([type(member)])
            if not isinstance(member, ModuleType):
                yield from _discover_cachers(member, objname + '.' + name, seen=seen)


class DiskCacher():
    '''Class implementing a persistent disk cache'''

    def __init__(self, f):
        self._tag = None
        self._data = None
        self._tmpdir = tempfile.gettempdir()
        self._original_function = f
        self._method_name = f.__qualname__
        self._funcid = None
        self._prefix = 'cache'

    def set_tag(self, tag, instance_name):
        '''Set the tag used to make cached data persistent'''
        self._tag = tag
        self._funcid = instance_name + '.' + self._method_name

    def set_tmpdir(self, tmpdir):
        '''Set the directory where cached data is stored'''
        self._tmpdir = tmpdir

    def set_prefix(self, prefix):
        '''Set prefix for tag directories'''
        self._prefix = prefix

    def clear_cache(self):
        '''Remove cache from disk and memory'''
        self._data = None
        if self._tag:
            self._delete_from_disk()

    def _log(self, message):
        if _logger:
            _logger.debug(' '.join((self._tag, self._funcid, message)))

    def execute(self, *args, **kwargs):
        '''Cache lookup'''
        if self._tag is not None:
            # Local memory cache
            if self._data is not None:
                self._log('Internal cache hit')
                return self._data
            # Disk cache
            try:
                self._data = self._load_from_disk()
                return self._data
            except FileNotFoundError as e:
                self._log('Exception reading from disk: '+str(e))
                self._log('Calling function to cache')
                self._data = self._original_function(*args, **kwargs)
                self._save_to_disk()
                return self._data
        else:
            return self._original_function(*args, **kwargs)

    def fullpath(self):
        '''Full path of cache file on disk, without extension'''
        if self._tag is None:
            raise TagNotSetException('Disk cache tag has not been set')
        if self._funcid is None:
            raise TagNotSetException('Function ID has not been set')
        return os.path.join(self._tmpdir, self._prefix + self._tag, self._funcid + '.npy')

    def _save_to_disk(self):
        self._log('Saving ' + self.fullpath())
        try:
            os.mkdir(os.path.join(os.path.dirname(self.fullpath())))
        except FileExistsError:
            pass

        try:
            # Use exact type match because e.g. u.Quantity is a subclass
            # but cannot be used with np.save
            if type(self._data) == np.ndarray:          # pylint: disable=C0123
                np.save(self.fullpath(), self._data)
            else:
                pickle.dump(self._data, open(self.fullpath(), 'wb'))
        except Exception as e:
            self._log('Exception saving to disk: '+str(e))

    def _load_from_disk(self):
        self._log('Reading ' + self.fullpath())
        return np.load(self.fullpath(), allow_pickle=True)

    def _delete_from_disk(self):
        self._log('Deleting ' + self.fullpath())
        if os.path.exists(self.fullpath()):
            os.unlink(self.fullpath())

# __oOo__




import abc
import io
import logging
import os
import re
import shlex
import subprocess
import sys
import time
import traceback
import unittest


ENV_VAR_LONG_RUNNING_TESTS_ENABLE = "ENABLE_LONG_RUNNING_TESTS"
XVFB_DEFAULT_DISPLAY = ":42"


def setUpLogger(loggingLevel):
    FORMAT = '%(asctime)s %(levelname)s %(name)s %(message)s'
    f = logging.Formatter(fmt=FORMAT)
    root_logger = logging.getLogger()
    root_logger.setLevel(loggingLevel)
    if not root_logger.handlers:
        handler = logging.StreamHandler()
        handler.setFormatter(f)
        handler.setLevel(loggingLevel)
        root_logger.addHandler(handler)


class TestHelper(object):

    TIMEOUT_FACTOR = 1.0

    @staticmethod
    def sleep(durationSec):
        time.sleep(durationSec)

    @staticmethod
    def longRunningTest(f):
        def wrappedMethod(self, *args, **kwds):
            if TestHelper.areLongRunningTestsInhibited():
                TestHelper._logSkippedTest()
                return
            else:
                return f(self, *args, **kwds)

        return wrappedMethod

    @staticmethod
    def areLongRunningTestsInhibited():
        return ENV_VAR_LONG_RUNNING_TESTS_ENABLE not in os.environ

    @staticmethod
    def _logSkippedTest():
        logging.warning("Test is skipped because long running tests"
                        " are inhibited.  Set environment variable %s"
                        " in order to enable this test!" % (
                            ENV_VAR_LONG_RUNNING_TESTS_ENABLE))

    @staticmethod
    def areDictEqual(d1, d2):
        items1 = list(d1.items())
        items2 = list(d2.items())
        if len(items1) != len(items2):
            return False

        for each in d1:
            if d1[each] != d2[each]:
                return False

        return True

    @staticmethod
    def whenLongRunningTestsAreInhibitedThenExitWithSuccess():
        if TestHelper.areLongRunningTestsInhibited():
            TestHelper._logSkippedTest()
            sys.exit(os.EX_OK)

    @staticmethod
    def checkThatFileContainsString(logFilePath, expectedMessage,
                                    timeoutInSec=5.0):
        WAIT_END = time.time() + timeoutInSec
        while not TestHelper.isFileContainingString(expectedMessage,
                                                    logFilePath):
            time.sleep(0.05)
            if time.time() > WAIT_END:
                raise RuntimeError(("File %s is lacking string '%s' "
                                    "after waiting for %.3f seconds" % (
                                        logFilePath, expectedMessage,
                                        timeoutInSec)))

    @staticmethod
    def isFileContainingString(expectedMessage, filePath):
        try:
            return 1 <= TestHelper.getNumberOfMatches(
                re.escape(expectedMessage)).inFile(filePath)
        except:
            return False

    @staticmethod
    def writeLinesToFile(filePath, fileContentsAsLines):
        f = open(filePath, "wb")
        for each in fileContentsAsLines:
            f.write(each)
            f.write('\n')
        f.close()

    @staticmethod
    def loggingTurnedOffMessageForPort(port):
        return "Logging is turned off for setting output port %d" % port

    @staticmethod
    def startService(serviceCommand, serviceLogPath, startUpMessage):
        assert 0 < len(serviceCommand)
        assert 0 < len(serviceLogPath)

        return TestHelper.startServiceWithArgs(
            shlex.split(serviceCommand),
            serviceLogPath,
            startUpMessage)

    @staticmethod
    def startServiceWithArgs(arguments, serviceLogPath, startUpMessage,
                             timeoutSec=7):
        assert len(arguments) > 0
        assert len(serviceLogPath) > 0

        serviceLog = open(serviceLogPath, "wb")
        try:
            proc = subprocess.Popen(arguments,
                                    stdout=serviceLog,
                                    stderr=serviceLog)
        except Exception as e:
            raise Exception("Failed to execute '%s' (%s)" % (
                            arguments, str(e)))

        Poller(timeoutSec, pollingDelaySec=0.2).check(
            MessageInFileProbe(startUpMessage, serviceLogPath))

        return proc

    @staticmethod
    def isSubprocessAlive(process):
        return process.poll() is None

    @staticmethod
    def terminateSubprocess(process):
        exitCode = process.poll()
        if exitCode is not None:
            logging.info("Process with PID %d already dead" % process.pid)
            return
        TestHelper._terminateByUsingPid(process)
        END_TIME = time.time() + 5.0
        done = False
        while not done:
            if process.poll() is not None:
                done = True
                break
            elif time.time() > END_TIME:
                assert False, "Process with PID %d failed to die" % process.pid

        assert process.returncode is not None

    @staticmethod
    def _terminateByUsingPid(process):
        starterPid = process.pid
        if os.EX_OK == subprocess.call("kill -INT %d" % starterPid,
                                       shell=True):
            try:
                Poller(5).check(SubprocessTerminated(process, "<no name>"))
                return
            except:
                pass

    @staticmethod
    def dumpFileToStdout(filePath):
        assert 0 < len(filePath)
        if not os.path.exists(filePath):
            print("File %s is missing" % filePath)
        else:
            print("--- BEGIN %s ---" % filePath)
            with io.open(filePath, "r") as f:
                for eachLine in f:
                    print(eachLine, end=' ')
            print("--- END %s ---" % filePath)
        sys.stdout.flush()

    @staticmethod
    def executeShellCommand(cmd):
        exitCode = subprocess.call(cmd, shell=True)
        if os.EX_OK != exitCode:
            raise Exception("Failed to execute shell command '%s'."
                            "  Exit code is: %d" % (cmd, exitCode))

    @staticmethod
    def removeFileIfAny(path):
        if os.path.exists(path):
            os.remove(path)

    @staticmethod
    def getNumberOfMatches(pattern):
        class MatchCounter(object):
            def __init__(self, pattern):
                self._pattern = pattern

            def inFile(self, filePath):
                with io.open(filePath, 'r') as fd:
                    content = fd.read()
                res = re.findall(self._pattern, content)
                return len(res)

        return MatchCounter(pattern)

    @staticmethod
    def getNumberOfMessages(message):
        return TestHelper.getNumberOfMatches(re.escape(message))

    @staticmethod
    def executeQtGUITestsWithXvfb(testClass, argv, XvfbLogPath='./xvfb.log'):
        result = None
        qtApp = None
        from PyQt5 import QtWidgets
        with Xvfb.launch(XvfbLogPath) as xvfb:
            qtApp = QtWidgets.QApplication(argv)
            result = unittest.TextTestRunner().run(
                unittest.TestLoader().loadTestsFromTestCase(testClass))
            assert qtApp is not None
            del qtApp

        assert result is not None
        if not result.wasSuccessful():
            sys.exit(os.EX_SOFTWARE)

    @staticmethod
    def qtPoll(timeoutSec):
        return _QtPollCheck(timeoutSec)


class _QtPollCheck(object):

    def __init__(self, timeoutSec):
        self._timeoutSec = timeoutSec

    def check(self, f, *args, **kwds):
        def _check():
            from PyQt5.Qt import QApplication
            QApplication.processEvents()
            f(*args, **kwds)

        Poller(self._timeoutSec).check(ExecutionProbe(_check))
        return self


def _pathRelativeToThisFile(path):
    return os.path.join(os.path.abspath(os.path.dirname(__file__)), path)


class Probe(object, metaclass=abc.ABCMeta):

    @abc.abstractmethod
    def probe(self):
        assert False, "Abstract method called"

    @abc.abstractmethod
    def isSatisfied(self):
        assert False, "Abstract method called"

    @abc.abstractmethod
    def errorMessage(self):
        assert False, "Abstract method called"


class Poller(object):

    def _adjustTimeoutIfComputerIsSlow(self):
        if TestHelper.TIMEOUT_FACTOR != 1.0:
            adjustedTimeoutSec = (
                self._timeoutSec * TestHelper.TIMEOUT_FACTOR)
            logging.info(
                "Increasing polling timeout from %.1f to %.1f sec"
                " because computer performance factor is set to %.2f." % (
                    self._timeoutSec, adjustedTimeoutSec,
                    TestHelper.TIMEOUT_FACTOR))
            self._timeoutSec = adjustedTimeoutSec

    def __init__(self, timeoutSec, pollingDelaySec=0.01, reportPeriodSec=1.0):
        assert timeoutSec >= 0
        self._timeoutSec = timeoutSec
        self._pollingDelaySec = pollingDelaySec
        self._reportPeriodSec = reportPeriodSec
        self._nextReportTimeSec = None
        self._t0 = None

        self._adjustTimeoutIfComputerIsSlow()

    def _reportProgressIfAny(self):
        now = time.time()
        if self._nextReportTimeSec is None:
            self._nextReportTimeSec = now + self._reportPeriodSec

        if time.time() > self._nextReportTimeSec:
            logging.info("Polling for %.1f sec.  Timeout is %.1f sec" % (
                now - self._t0, self._timeoutSec))
            self._nextReportTimeSec += self._reportPeriodSec

    def check(self, probe):
        self._t0 = time.time()
        END_TIME_SEC = self._t0 + self._timeoutSec
        done = False
        while not done:
            self._reportProgressIfAny()
            probe.probe()
            if probe.isSatisfied():
                done = True
            elif time.time() >= END_TIME_SEC:
                raise RuntimeError(
                    "Timeout of %s seconds while polling (%s)" % (
                        self._timeoutSec, probe.errorMessage()))
            else:
                time.sleep(self._pollingDelaySec)


class CommandOutputProbe(Probe):

    def __init__(self, shellCommand, expectedStringInOutput):
        Probe.__init__(self)
        self._shellCmd = shellCommand
        self._expectedString = expectedStringInOutput
        self._satisfied = False

    def probe(self):
        process = subprocess.Popen(shlex.split(self._shellCmd),
                                   stdout=subprocess.PIPE)
        out, _ = process.communicate()
        if self._expectedString in out:
            self._satisfied = True

    def isSatisfied(self):
        return self._satisfied

    def errorMessage(self):
        return "Failed to find '%s' in output of command '%s'" % (
               self._expectedString, self._shellCmd)


class SuccessfullShellCommandProbe(Probe):

    def __init__(self, command):
        Probe.__init__(self)
        self._command = command

    def probe(self):
        self._exitCode = os.system(self._command)

    def isSatisfied(self):
        return os.EX_OK == self._exitCode

    def errorMessage(self):
        return "Failed to run '%s'" % (self._command)


class MessageInFileProbe(Probe):

    def __init__(self, expectedMessage, filePath):
        self._expectedMessage = expectedMessage
        self._filePath = filePath

    def probe(self):
        pass

    def isSatisfied(self):
        return TestHelper.isFileContainingString(self._expectedMessage,
                                                 self._filePath)

    def errorMessage(self):
        return "Message '%s' is missing in file %s" % (self._expectedMessage,
                                                       self._filePath)


class SubprocessTerminated(Probe):

    def __init__(self, process, name):
        self._process = process
        self._name = name

    def probe(self):
        self._process.poll()

    def isSatisfied(self):
        return self._process.returncode is not None

    def errorMessage(self):
        return "Subprocess '%s' is not terminated" % self._name


class ExecutionProbe(Probe):

    def __init__(self, func, name="", *args, **kwds):
        self._func = func
        self._succeeded = False
        self._name = name
        self._args = args
        self._kwds = kwds
        self._lastException = None
        self._lastTraceback = None

    @staticmethod
    def create(func, *args, **kwds):
        return ExecutionProbe(func, "no_name", *args, **kwds)

    def probe(self):
        try:
            self._func(*self._args, **self._kwds)
            self._succeeded = True
        except Exception as e:
            self._lastException = e
            self._lastTracebackFormat = traceback.format_exc()
            logging.debug(e)

    def isSatisfied(self):
        return self._succeeded

    def errorMessage(self):
        return ("Failed to execute function %s (%s). "
                "Last exception: '%s' (%s)") % (
                    self._func.__name__, self._name, str(self._lastException),
                    self._lastTracebackFormat)


class PollingTestTask(object):

    def __init__(self, probe, timeoutSec):
        self._probe = probe
        self._timeoutSec = timeoutSec
        self._startTime = None
        self._waitEndTime = None

    def hasPassed(self):
        return self._probe.isSatisfied()

    def check(self):
        if self._startTime is None:
            self._startTime = time.time()
            self._waitEndTime = self._startTime + self._timeoutSec

        self._probe.probe()
        if self._probe.isSatisfied():
            pass
        elif time.time() >= self._waitEndTime:
            raise RuntimeError(
                "Timeout of %s seconds occured while polling (%s)" % (
                    self._timeoutSec, self._probe.errorMessage()))


class TestSuiteTask(object):

    def __init__(self, testTasks):
        self._testTasks = testTasks

    def check(self):
        try:
            self._check()
        except Exception as e:
            logging.error(str(e))
            sys.exit(1)

    def _check(self):
        nPassedTasks = 0
        for each in self._testTasks:
            each.check()
            if each.hasPassed():
                nPassedTasks += 1

        totalTests = len(self._testTasks)
        if totalTests == nPassedTasks:
            logging.info("%d of %d tests have passed" % (
                nPassedTasks, totalTests))
            sys.exit(os.EX_OK)


def terminateSubprocess(process):
    TestHelper.terminateSubprocess(process)


def isSubprocessAlive(process):
    return TestHelper.isSubprocessAlive(process)


def dumpFileToStdout(filePath):
    TestHelper.dumpFileToStdout(filePath)


class Xvfb(object):

    XVFB_PATH = "/usr/bin/Xvfb"
    X_SERVER_CHECK_COMMAND_LOG_PATH = "./xdpyinfo.log"
    X_SERVER_CHECK_COMMAND = "xdpyinfo >%s 2>&1" % (
        X_SERVER_CHECK_COMMAND_LOG_PATH)

    @staticmethod
    def launch(xvfbLogFilePath):
        os.environ["DISPLAY"] = XVFB_DEFAULT_DISPLAY
        xvfb = Xvfb(xvfbLogFilePath, XVFB_DEFAULT_DISPLAY)
        xvfb.start()

        return xvfb

    def __init__(self, xvfbLogFilePath, displayId):
        self._xvfbLogFilePath = xvfbLogFilePath
        self._displayId = displayId

        if not self._isExe(Xvfb.XVFB_PATH):
            logging.error("Xvfb does not exist in '%s'!" % Xvfb)
            sys.exit(1)

    def _isExe(self, fpath):
        return os.path.isfile(fpath) and os.access(fpath, os.X_OK)

    def start(self):
        self._xvfbLogFile = open(self._xvfbLogFilePath, "wb")
        xvfbStartCommand = "%s %s -screen 0 1024x768x24" % (
            Xvfb.XVFB_PATH, self._displayId)
        self._xvfb = subprocess.Popen(
            shlex.split(xvfbStartCommand),
            stdout=self._xvfbLogFile, stderr=self._xvfbLogFile)
        self._waitUntilXvfbIsReady()

    def _waitUntilXvfbIsReady(self):

        class XvfbStartupProbe(Probe):

            def __init__(self, displayId):
                self._displayId = displayId
                self._done = False

            def probe(self):
                assert self._displayId == os.environ["DISPLAY"]
                exitCode = os.system(Xvfb.X_SERVER_CHECK_COMMAND)
                if os.EX_OK == exitCode:
                    self._done = True

            def isSatisfied(self):
                return self._done

            def errorMessage(self):
                return "Failed to detect Xvfb startup"

        try:
            Poller(timeoutSec=5.0).check(XvfbStartupProbe(self._displayId))
        except:
            TestHelper.dumpFileToStdout(self.X_SERVER_CHECK_COMMAND_LOG_PATH)
            raise

    def stop(self):
        terminateSubprocess(self._xvfb)
        self._xvfbLogFile.close()
        os.remove(self._xvfbLogFilePath)

    def __enter__(self):
        return self

    def __exit__(self, type, value, traceback):
        self.stop()



import abc
from astropy.io import fits as pyfits
from six import with_metaclass


class Snapshotable(with_metaclass(abc.ABCMeta, object)):
    @abc.abstractmethod
    def get_snapshot(self, prefix):
        assert False

    @staticmethod
    def _isSuitableForFITSHeader(value):
        return value is not None

    @staticmethod
    def _truncateStringIfAny(key, value):
        maxValueLenInChars = 67 - len(key)
        if len(value) > maxValueLenInChars:
            return value[0:maxValueLenInChars]
        return value

    @staticmethod
    def _updateHeader(hdr, key, value):
        MAX_KEY_LEN_CHARS = 59
        assert len(key) <= MAX_KEY_LEN_CHARS
        if isinstance(value, str):
            value = Snapshotable._truncateStringIfAny(key, value)
        hdr.update({'hierarch ' + key: value})

    @staticmethod
    def as_fits_header(snapshotDictionary):
        hdr = pyfits.Header()
        for k in sorted(snapshotDictionary.keys()):
            value = snapshotDictionary[k]
            if Snapshotable._isSuitableForFITSHeader(value):
                Snapshotable._updateHeader(hdr, k, value)
        return hdr

    @staticmethod
    def prepend(prefix, snapshotDict):
        assert len(prefix) > 0, "Prefix length must be greater than zero"
        for each in list(snapshotDict.keys()):
            value = snapshotDict[each]
            del snapshotDict[each]
            newKey = prefix + "." + each
            snapshotDict[newKey] = value
        return snapshotDict

    @staticmethod
    def from_fits_header(hdr):
        snapshot = {}
        for each in hdr:
            snapshot[each] = hdr[each]
        return snapshot

    @staticmethod
    def remove_entries_with_value_none(snapshotDict):
        for each in list(snapshotDict.keys()):
            if snapshotDict[each] is None:
                del snapshotDict[each]




"""
Display functions
"""

import numpy as np


def movie(frames, interval=0.1, *args, **kwargs):
    '''
    Parameters
    ----------
    frames: ndarray
        3d array [time, rows, cols] with the data to display
    interval: float. optional
        delay between frames in seconds, default=0.1 seconds
    '''
    import matplotlib.pyplot as plt
    fig, ax = plt.subplots(1, 1)

    for i, frame in enumerate(frames):
        if i == 0:
            img = ax.imshow(frame, *args, **kwargs)
        else:
            img.set_data(frame)
        fig.canvas.draw_idle()
        plt.pause(interval)


def savegif(frames, filename, interval=0.1, loop=0):
    '''
    Parameters
    ----------
    frames: ndarray
        3d array [time, rows, cols] with the data to save
    filename: string, pathlib.Path object or file object
        filename or path or fileobject where the animated GIF is saved
    interval: float
        delay between frames in seconds
    loop: int
        number of times the GIF should loop. Default 0, which means to loop forever.
    '''
    from PIL import Image
    imgs = [Image.fromarray(img) for img in frames]
    imgs[0].save(filename, format='gif', save_all=True, append_images=imgs[1:], duration=interval, loop=loop)


def _grouper(iterable, n):
    # From itertools recipe: grouper('ignore'):
    # grouper('ABCDEFG', 3) --> ABC DEF
    return zip(*[iter(iterable)] * n)


def tile(frames, rowlength=10):
    '''
    Parameters
    ----------
    frames: ndarray
        3d array [time, rows, cols] with the data to display
    rowlength: int, optional
        number of frames to arrange in a row
    '''
    rows = [np.hstack(x) for x in _grouper(frames, rowlength)]
    return np.vstack(rows)


# ___oOo___



from arte.utils.package_data import dataRootDir
import os
from synphot.spectrum import SpectralElement
from astropy import units as u


class Filters(object):

    BESSELL_J = 'bessel_j'
    BESSELL_H = 'bessel_h'
    BESSELL_K = 'bessel_k'
    BESSELL90_R = 'bessell90_r'
    BESSELL90_I = 'bessell90_i'
    COUSINS_R = 'cousins_r'
    COUSINS_I = 'cousins_i'
    JOHNSON_U = 'johnson_u'
    JOHNSON_B = 'johnson_b'
    JOHNSON_V = 'johnson_v'
    JOHNSON_R = 'johnson_r'
    JOHNSON_I = 'johnson_i'
    JOHNSON_J = 'johnson_j'
    JOHNSON_K = 'johnson_k'
    ESO_ETC_U = 'eso_etc_u'
    ESO_ETC_B = 'eso_etc_b'
    ESO_ETC_V = 'eso_etc_v'
    ESO_ETC_R = 'eso_etc_r'
    ESO_ETC_I = 'eso_etc_i'
    ESO_ETC_Z = 'eso_etc_z'
    ESO_ETC_Y = 'eso_etc_y'
    ESO_ETC_J = 'eso_etc_j'
    ESO_ETC_H = 'eso_etc_h'
    ESO_ETC_K = 'eso_etc_k'
    ESO_ETC_L = 'eso_etc_l'
    ESO_ETC_M = 'eso_etc_m'
    ESO_ETC_N = 'eso_etc_n'
    ESO_ETC_Q = 'eso_etc_q'
    C_RED_ONE = 'c_red_one'
    CCD_220 = 'ccd_220'

    SYNPHOT = [
        BESSELL_J,
        BESSELL_H,
        BESSELL_K,
        COUSINS_R,
        COUSINS_I,
        JOHNSON_U,
        JOHNSON_B,
        JOHNSON_V,
        JOHNSON_R,
        JOHNSON_I,
        JOHNSON_J,
        JOHNSON_K
    ]

    ESO_ETC = [
        ESO_ETC_U,
        ESO_ETC_B,
        ESO_ETC_V,
        ESO_ETC_R,
        ESO_ETC_I,
        ESO_ETC_Z,
        ESO_ETC_Y,
        ESO_ETC_J,
        ESO_ETC_H,
        ESO_ETC_K,
        ESO_ETC_L,
        ESO_ETC_M,
        ESO_ETC_N,
        ESO_ETC_Q
    ]

    BESSELL90 = [
        BESSELL90_R,
        BESSELL90_I
    ]

    ALL = SYNPHOT + ESO_ETC + BESSELL90

    @classmethod
    def names(cls):
        return cls.ALL

    @classmethod
    def get(cls, filter_name):
        '''
        Return synphot.SpectralElement of the specified filter
        The list of available filter is accessible via Filters.names()
        '''
        if filter_name in cls.SYNPHOT:
            return SpectralElement.from_filter(filter_name)
        else:
            method_to_call = getattr(cls, '_%s' % filter_name)
            return method_to_call()

    @staticmethod
    def _EsoEtcFiltersFolder():
        rootDir = dataRootDir()
        dirname = os.path.join(rootDir, 'photometry', 'filters', 'eso_etc')
        return dirname

    @staticmethod
    def _Bessell90FiltersFolder():
        rootDir = dataRootDir()
        dirname = os.path.join(rootDir, 'photometry', 'filters', 'bessell90')
        return dirname

    @staticmethod
    def _DetectorsFolder():
        rootDir = dataRootDir()
        dirname = os.path.join(rootDir, 'photometry', 'detectors')
        return dirname

    @classmethod
    def _eso_etc_u(cls):
        return cls._eso_etc('u', u.nm)

    @classmethod
    def _eso_etc_b(cls):
        return cls._eso_etc('b', u.nm)

    @classmethod
    def _eso_etc_v(cls):
        return cls._eso_etc('v', u.nm)

    @classmethod
    def _eso_etc_r(cls):
        return cls._eso_etc('r', u.nm)

    @classmethod
    def _eso_etc_i(cls):
        return cls._eso_etc('i', u.nm)

    @classmethod
    def _eso_etc_z(cls):
        return cls._eso_etc('z', u.um)

    @classmethod
    def _eso_etc_y(cls):
        return cls._eso_etc('y', u.nm)

    @classmethod
    def _eso_etc_j(cls):
        return cls._eso_etc('j', u.um)

    @classmethod
    def _eso_etc_h(cls):
        return cls._eso_etc('h', u.um)

    @classmethod
    def _eso_etc_k(cls):
        return cls._eso_etc('k', u.um)

    @classmethod
    def _eso_etc_l(cls):
        return cls._eso_etc('l', u.um)

    @classmethod
    def _eso_etc_m(cls):
        return cls._eso_etc('m', u.um)

    @classmethod
    def _eso_etc_n(cls):
        return cls._eso_etc('n', u.um)

    @classmethod
    def _eso_etc_q(cls):
        return cls._eso_etc('q', u.um)

    @classmethod
    def _ccd_220(cls):
        return SpectralElement.from_file(
            os.path.join(cls._DetectorsFolder(), 'ccd220.dat'),
            wave_unit=u.um)

    @classmethod
    def _c_red_one(cls):
        return SpectralElement.from_file(
            os.path.join(cls._DetectorsFolder(), 'c_red_one.dat'),
            wave_unit=u.nm)

    @classmethod
    def _eso_etc(cls, filter_name, wavelength_unit):
        return SpectralElement.from_file(
            os.path.join(cls._EsoEtcFiltersFolder(),
                         'phot_%s.dat' % filter_name),
            wave_unit=wavelength_unit)

    @classmethod
    def _bessell90(cls, filter_name):
        return SpectralElement.from_file(
            os.path.join(cls._Bessell90FiltersFolder(),
                         'bessell90_%s.dat' % filter_name),
            wave_unit=u.AA)

    @classmethod
    def _bessell90_r(cls):
        return cls._bessell90('r')

    @classmethod
    def _bessell90_i(cls):
        return cls._bessell90('i')



# -*- coding: utf-8 -*-
#########################################################
#
# who       when        what
# --------  ----------  ---------------------------------
# apuglisi  2012-11-05  Created
#
#########################################################
import os
import fnmatch


def locate(pattern, rootdir=None):
    '''Generator similar to Unix's *locate* utility

    Locates all files matching a pattern, inside and below the root directory.
    If no root directory is given, the current directory is used instead.

    Parameters
    ----------
    pattern: string
        the filename pattern to match. Unix wildcards are allowed
    rootdir: string, optional
        the root directory where the search is started. If not set,
        the current directory is used.

    Yields
    ------
    string
        the next filename matching *pattern*

    '''
    if rootdir is None:
        rootdir = os.curdir

    for path, dirs, files in os.walk(rootdir):
        for filename in fnmatch.filter(files, pattern):
            yield os.path.join(path, filename)


def locate_first(pattern, rootdir=None):
    '''Locate the first filename matching *pattern*

    Locate the first file a matching, inside and below the root directory.
    If no root directory is given, the current directory is used instead.

    Parameters
    ----------
    pattern: string
        the filename pattern to match. Unix wildcards are allowed
    rootdir: string, optional
        the root directory where the search is started. If not set,
        the current directory is used.

    Returns
    -------
    string
        the first filename matching *pattern*, or None if not found.
    '''
    loc = locate(pattern, rootdir)
    try:
        return next(loc)
    except StopIteration:
        return None


def replace_in_file(filename, search, replace):
    '''Replaces a string inside a file'''

    filedata = None
    with open(filename, 'r') as f:
        filedata = f.read()

    # Replace the target string
    filedata = filedata.replace(search, replace)

    # Write the file out again
    with open(filename, 'w') as f:
        f.write(filedata)

# ___oOo___



import numpy as np
from arte.utils.zernike_generator import ZernikeGenerator


class NoisePropagationZernikeGradientWFS():

    def __init__(self,
                 nSubaps=3,
                 modesV=np.arange(2, 7),
                 rcond=0.1):
        self._nSubaps= nSubaps
        self._zg= ZernikeGenerator(nSubaps)
        self._modesV= modesV
        self._rcond= rcond
        self._initialize()
        
    def _initialize(self):
        self._phi = None
        self._D = None 
        self._u = None 
        self._s = None 
        self._vh = None
        self._sinv = None  
        self._R = None 
        

    def _compute_interaction_matrix(self):
        self._phi= np.dstack([
            self._zg.getZernike(m).filled(0) for m in self._modesV])
        self._D= np.vstack(
            [np.hstack(
                (self._zg.getDerivativeX(m).compressed(),
                 self._zg.getDerivativeY(m).compressed()))
             for m in self._modesV]).T

    def _compute_reconstructor(self):
        self._u, self._s, self._vh= np.linalg.svd(
            self.interaction_matrix, full_matrices=False)
        self._sinv= 1/ self._s
        self._sinv[np.argwhere(self._s< self._rcond * self._s.max())]=0
        self._R=np.matmul(self._vh.T,
                          np.matmul(np.diag(self._sinv), self._u.T))

    @property
    def nSubaps(self):
        return self._nSubaps


    @property
    def nSlopes(self):
        return self.interaction_matrix.shape[0]


    @property
    def modesVector(self):
        return self._modesV


    @property
    def nModes(self):
        return self.interaction_matrix.shape[1]


    @property
    def phaseCube(self):
        return self._phi


    @property
    def reconstructor(self):
        if self._R is None:
            self._compute_reconstructor()
        return self._R


    @property
    def interaction_matrix(self):
        if self._D is None:
            self._compute_interaction_matrix()
        return self._D


    @property
    def s(self):
        if self._s is None:
            self._compute_reconstructor()
        return self._s


    @property
    def sInv(self):
        if self._sinv is None:
            self._compute_reconstructor()
        return self._sinv


    @property
    def u(self):
        if self._u is None:
            self._compute_reconstructor()
        return self._u


    @property
    def v(self):
        if self._vh is None:
            self._compute_reconstructor()
        return self._vh.T


    @property
    def noise_covariance_matrix(self):
        return np.dot(self.reconstructor, self.reconstructor.T)


    @property
    def noise_total_variance(self):
        return np.trace(self.noise_covariance_matrix)


    def noiseSimulation(self, sigma=1, nSamples=1000):
        return np.var(np.dot(
            self.reconstructor, sigma * np.random.randn(self.nSlopes, nSamples)), axis=1)


    def _noise_rg(self, radial_order):
        return -2.05 * np.log10(radial_order + 1) - 0.53


    def _noise_rg2(self, radial_order):
        return -2.0 * np.log10(radial_order + 1) - 0.76



    # def slopesMapForMode(self, nMode):
    #     return self.interaction_matrix[:, nMode].filled(0).reshape(
    #         self.nSubaps* 2, self.nSubaps)
    #
    #
    # def leftSingularVectorMapForMode(self, nMode):
    #     return self.u[:, nMode].reshape(
    #         self.nSubaps* 2, self.nSubaps)



# -*- coding: utf-8 -*-

import numpy as np
from arte.utils.help import add_help, modify_help
from arte.utils.shared_array import SharedArray


@add_help
class GenericCircularBuffer():
    '''
    Generic circular buffer.

    Implements a generic circular buffer using any class with a
    numpy-like interface (must define a constructor with shape and dtype
    parameters, and the constructed object must have a read/write [] operator)
    '''

    def __init__(self, n_frames, shape, dtype, constructor):
        self._buf = constructor((n_frames,) + shape, dtype=dtype)
        self._len = n_frames
        self._counter = constructor((1,), dtype=np.uint32)
        self._position = constructor((1,), dtype=np.uint32)

    @modify_help(arg_str='data, position=None')
    def store(self, data, position=None):
        '''
        Store a record in the circular buffer.

        By default, the record is stored following an internal counter,
        which is then incremented.
        '''
        if position is None:
            position = self._position[0]

        self._buf[position,:] = data
        self._counter[0] = self._counter[0] + 1
        self._position[0] = self._counter[0] % self._len

    def get(self, position):
        '''Returns a frame from the circular buffer'''
        return self._buf[position,:]

    def position(self):
        '''Returns the current position in the circular buffer'''
        return self._position[0]

    def counter(self):
        '''Returns the total number of stored frames'''
        return self._counter[0]


class NumpyCircularBuffer(GenericCircularBuffer):
    '''Implements a circular buffer on top of a numpy array'''

    def __init__(self, n_frames, shape, dtype):
        super().__init__(n_frames, shape, dtype, np.zeros)


class SharedCircularBuffer(GenericCircularBuffer):
    '''
    Implements a circular buffer on top of a SharedArray

    .. warning:: Since 3.8 the Python standard library provides a SharedMemory
                 class that makes this implementation not working.

    .. warning:: It does not work on Windows.

    '''

    def __init__(self, n_frames, shape, dtype):
        super().__init__(n_frames, shape, dtype, SharedArray)





class CircularRange():
    '''Circular range()-like object.'''

    def __init__(self, start, stop, modulo):
        self.start = start
        self.stop = stop
        self.modulo = modulo
        self.index = start

    def __repr__(self):
        return 'CircularRange(%d, %d, %d)' % (self.start, self.stop, self.modulo)

    def __iter__(self):
        while self.index != self.stop:
            yield self.index
            self.index = (self.index + 1) % self.modulo





import numpy as np


def show_array(array, cut_wings=0, title=None, xlabel='column', ylabel='row', units=None, **kwargs):
    '''
    pyplot.matshow() wrapper to add a colorbar and title/labels

    cut_wings=x means that colorbar is saturated for array values below x percentile
    and above 100-x percentile. Default is 0, i.e. all data are displayed; values below
    0 are forced to 0, values above 50 are set to 50.
    '''
    import matplotlib.pyplot as plt

    if cut_wings <= 0:
        imgplt = plt.matshow(array, **kwargs)
        clb = plt.colorbar()
    else:
        if cut_wings >= 50:
            tmp_cut_wings = 50
        else:
            tmp_cut_wings = cut_wings
        
        vmin = np.percentile(array, tmp_cut_wings)
        vmax = np.percentile(array, 100-tmp_cut_wings)
        imgplt = plt.matshow(array, vmin=vmin, vmax=vmax, **kwargs)
        clb = plt.colorbar(extend='both')
    
    if units is not None:
        clb.ax.set_title(units)
    
    if title is not None:
        imgplt.axes.set_title(title)
    imgplt.axes.set_xlabel(xlabel)
    imgplt.axes.set_ylabel(ylabel)
    plt.show()
    
    return imgplt
 



import collections


class _AnalyzerStub(list):

    def __getattr__(self, name):
        # Necessary to allow casting by np.array()
        if name.startswith('__array'):
            raise AttributeError

        if hasattr(self[0], name):
            return _AnalyzerStub([getattr(x, name) for x in self])
        else:
            raise AttributeError

    def __call__(self, *args, **kwargs):
        return _AnalyzerStub([x(*args, **kwargs) for x in self])


class BaseAnalyzerSet():
    '''
    Analyzer set. Holds a list of Analyzer objects
    '''
    def __init__(self, from_or_list, to=None, recalc=False, *,
                       file_walker, analyzer_type):
        '''
        This constructor only builds the list of tags but does not allocate any Analyzer.

        Parameters
        ----------
        from_or_list: str or list
            either a single tag, or a list of tags
        to: str, optional
            sigle tag, or None
        recalc: bool
            if True, all analyzers will be recalculated (lazy recalc, only when actually accessed)
        file_walker: BaseFileWalker or derived class instance
            file walker used to find tag data (required keyword argument)
        analyzer_type: BaseAnalyzer or derived class (not an instance).
            analyzer type to instance for each tag. Must define a get() classmethod (required keyword argument).
        '''
        self._analyzer_args = []
        self._analyzer_kwargs = {}
        self._file_walker = file_walker
        self._analyzer_type = analyzer_type

        if isinstance(from_or_list, collections.abc.Sequence) and not isinstance(from_or_list, str):
            self.tag_list = sorted(from_or_list)
        else:
            self.tag_list = sorted(self._file_walker.find_tag_between_dates(str(from_or_list), str(to)))
        self._init_recalcs = {k: recalc for k in self.tag_list}

    def remove_invalids(self):
        '''
        Remove tags that evaluate to NotAvailable when created
        '''
        newtags = []
        for tag in self.tag_list:
            ee = self.get(tag, recalc=self._init_recalcs[tag])
            if str(ee) != 'NA':
                newtags.append(tag)
        self.tag_list = newtags

    def set_analyzer_args(self, *args, **kwargs):
        '''Set the additional arguments to pass to Analyzers' contructors'''
        self._analyzer_args = args
        self._analyzer_kwargs = kwargs

    def __iter__(self):
        for tag in self.tag_list:
            yield self.get(tag)

    def get(self, tag, recalc=False):
        '''Returns the Analyzer instance for this tag'''
        my_recalc = self._init_recalcs[tag] or recalc
        self._init_recalcs[tag] = False
        return self._analyzer_type.get(tag, *self._analyzer_args,
                                    recalc=my_recalc, **self._analyzer_kwargs)

    def __getitem__(self, idx_or_tag):
        if isinstance(idx_or_tag, int):
            return self.get(self.tag_list[idx_or_tag])
        else:
            return self.get(idx_or_tag)

    def append(self, tag):
        self.tag_list.append(tag)

    def insert(self, idx, tag):
        self.tag_list.insert(idx, tag)

    def remove(self, tag):
        _= self.tag_list.remove(tag)

    def __len__(self):
        return len(self.tag_list)

    def _apply(self, func_name, *args, **kwargs):

        for tag in self.tag_list:
            getattr(self.get(tag), func_name).__call__(*args, **kwargs)

    def _apply_w_args(self, func_name, args_list, kwargs_list):

        for tag,args,kwargs in zip(self.tag_list, args_list, kwargs_list):
            getattr(self.get(tag), func_name).__call__(*args, **kwargs)

    def generate_tags(self):
        for tag in self.tag_list:
            yield self.get(tag)

    def __getattr__(self, attrname):
        if hasattr(self.get(self.tag_list[0]), attrname):
            return _AnalyzerStub([getattr(self.get(tag), attrname) for tag in self.tag_list])
        else:
            raise AttributeError

    def wiki(self):
        '''Print wiki info on stdout'''
        for i, ee in enumerate(self.generate_tags()):
            ee.wiki(header = (i == 0))

    




import ctypes
import numpy as np
import multiprocessing as mp
from functools import reduce

from arte.utils.help import add_help


@add_help
class SharedArray:
    '''
    Class for a numpy-like buffer built on top of multiprocessing.

    SharedArray instances can be passed as arguments to processes
    created with the multiprocessing module, and each of them can call the
    ndarray() method to get a local view of the array.

    No access synchronization is provided.

    As a shortcut to read/write the array, the [] operator is supported,
    thus these two statements are equivalent::

        >>> array.ndarray()[2] = 3.1415
        >>> array[2] = 3.1415

    .. warning:: SharedArray works with processes spawned with mp.Process,
                 but do not work with mp.Pool, unless an mp.Manager is used.

    .. warning:: Since 3.8 the Python standard library provides a SharedMemory
                 class that must be used instead of this.

    .. warning:: It does not work on Windows.

    Parameters
    ----------
    shape: integer sequence
         array shape
    type: numpy dtype
         array dtype
    '''

    def __init__(self, shape, dtype):

        self.shape = shape
        self.dtype = np.dtype(dtype)
        self._ndarray = None

        n_elements = reduce(lambda x, y: x * y, self.shape)
        n_bytes = n_elements * self.dtype.itemsize
        self._shared_buf = mp.sharedctypes.RawArray(ctypes.c_byte, n_bytes)

    def __getitem__(self, key):
        return self.ndarray()[key]

    def __setitem__(self, key, value):
        self.ndarray()[key] = value

    def ndarray(self, realloc=False):
        '''
        Returns a new numpy wrapper around the buffer contents.

        Call this function after a task has been spawned the multiprocessing
        module in order to have access to the shared memory segment.

        If the array had already been accessed before passing it to the
        multiprocessing task, the task has to set `realloc` to True
        in order to reallocate a local copy of the array.

        Parameters
        ----------
        Realloc: bool, optional
            force array reallocation. Defaults to False

        Returns
        -------
        numpy.ndarray
            the shared numpy array. The array is read-write and changes
            will be immediately visible to other processes.
        '''
        if (self._ndarray is None) or realloc:
            arr = np.frombuffer(self._shared_buf, dtype=self.dtype)
            self._ndarray = arr.reshape(self.shape)
        return self._ndarray
